from __future__ import print_function
import gc, os, sys
import numpy as np
import scipy as sp
import numpy.linalg as la
import scipy.linalg as sla
from numpy.linalg import norm
from time import time
from copy import deepcopy
from warnings import warn
from time import time

from Florence.FiniteElements.Assembly import Assemble
from Florence import Mesh
from Florence.PostProcessing import PostProcess
from .StructuralDynamicIntegrator import StructuralDynamicIntegrator

__all__ = ["NonlinearImplicitStructuralDynamicIntegrator", "LinearImplicitStructuralDynamicIntegrator"]


class NonlinearImplicitStructuralDynamicIntegrator(StructuralDynamicIntegrator):
    """Implicit dynamic solver for nonlinear problems based on Newmark's beta
    """

    def __init__(self):
        super(NonlinearImplicitStructuralDynamicIntegrator, self).__init__()
        self.gamma   = 0.5
        self.beta    = 0.25


    def Solver(self, function_spaces, formulation, solver,
        K, M, NeumannForces, NodalForces, Residual,
        mesh, TotalDisp, Eulerx, Eulerp, material, boundary_condition, fem_solver):


        # COMPUTE DAMPING MATRIX BASED ON MASS
        D = 0.0
        if fem_solver.include_physical_damping:
            D = fem_solver.damping_factor*M

        # GET BOUNDARY CONDITIONS INFROMATION
        if formulation.fields == "electro_mechanics":
            self.GetBoundaryInfo(mesh, formulation,boundary_condition)
            M_mech = M[self.mechanical_dofs,:][:,self.mechanical_dofs]
            if fem_solver.include_physical_damping:
                D_mech = D[self.mechanical_dofs,:][:,self.mechanical_dofs]

        # INITIALISE VELOCITY AND ACCELERATION
        velocities     = np.zeros((mesh.points.shape[0],formulation.ndim))
        accelerations  = np.zeros((mesh.points.shape[0],formulation.ndim))

        # COMPUTE INITIAL ACCELERATION FOR TIME STEP 0
        InitResidual = Residual - NeumannForces[:,0][:,None]
        if formulation.fields == "electro_mechanics":
            accelerations[:,:] = solver.Solve(M_mech, -InitResidual[self.mechanical_dofs].ravel()
                ).reshape(mesh.points.shape[0],formulation.ndim)
        else:
            accelerations[:,:] = solver.Solve(M, -InitResidual.ravel() ).reshape(mesh.points.shape[0],formulation.ndim)

        self.NRConvergence = fem_solver.NRConvergence
        LoadIncrement = fem_solver.number_of_load_increments
        LoadFactor = fem_solver.total_time/LoadIncrement
        AppliedDirichletInc = np.zeros(boundary_condition.applied_dirichlet.shape[0],dtype=np.float64)

        if NeumannForces.ndim == 2 and NeumannForces.shape[1]==1:
            tmp = np.zeros((NeumannForces.shape[0],LoadIncrement))
            tmp[:,0] = NeumannForces[:,0]
            NeumannForces = tmp


        # TIME LOOP
        for Increment in range(1,LoadIncrement):

            t_increment = time()

            # APPLY NEUMANN BOUNDARY CONDITIONS
            DeltaF = NeumannForces[:,Increment][:,None]
            NodalForces = DeltaF

            # OBRTAIN INCREMENTAL RESIDUAL - CONTRIBUTION FROM BOTH NEUMANN AND DIRICHLET
            Residual = -boundary_condition.ApplyDirichletGetReducedMatrices(K,Residual,
                boundary_condition.applied_dirichlet[:,Increment],LoadFactor=1.0,mass=M,only_residual=True)
            Residual -= DeltaF
            # Residual = -DeltaF
            # GET THE INCREMENTAL DIRICHLET
            AppliedDirichletInc = boundary_condition.applied_dirichlet[:,Increment]

            # COMPUTE INITIAL ACCELERATION - ONLY NEEDED IN CASES OF PRESTRETCHED CONFIGURATIONS
            # accelerations[:,:] = solver.Solve(M, Residual.ravel() - \
            #     K.dot(TotalDisp[:,:,Increment].ravel())).reshape(mesh.points.shape[0],formulation.nvar)

            # LET NORM OF THE FIRST RESIDUAL BE THE NORM WITH RESPECT TO WHICH WE
            # HAVE TO CHECK THE CONVERGENCE OF NEWTON RAPHSON. TYPICALLY THIS IS
            # NORM OF NODAL FORCES
            if Increment==1:
                self.NormForces = np.linalg.norm(Residual)
                # AVOID DIVISION BY ZERO
                if np.isclose(self.NormForces,0.0):
                    self.NormForces = 1e-14
            self.norm_residual = np.linalg.norm(Residual)/self.NormForces


            Eulerx, Eulerp, K, Residual, velocities, accelerations = self.NewtonRaphson(function_spaces, formulation, solver,
                Increment, K, D, M, NodalForces, Residual, mesh, Eulerx, Eulerp,
                material,boundary_condition,AppliedDirichletInc, fem_solver, velocities, accelerations)

            # UPDATE DISPLACEMENTS FOR THE CURRENT LOAD INCREMENT
            TotalDisp[:,:formulation.ndim,Increment] = Eulerx - mesh.points
            if formulation.fields == "electro_mechanics":
                TotalDisp[:,-1,Increment] = Eulerp

            # COMPUTE DISSIPATION OF ENERGY THROUGH TIME
            if fem_solver.compute_energy_dissipation:
                energy_info = self.ComputeEnergyDissipation(function_spaces[0],mesh,material,formulation,fem_solver,
                    Eulerx, TotalDisp, NeumannForces, M, velocities, Increment)
                formulation.energy_dissipation.append(energy_info[0])
                formulation.internal_energy.append(energy_info[1])
                formulation.kinetic_energy.append(energy_info[2])
                formulation.external_energy.append(energy_info[3])
            # COMPUTE DISSIPATION OF LINEAR MOMENTUM THROUGH TIME
            if fem_solver.compute_linear_momentum_dissipation:
                power_info = self.ComputePowerDissipation(function_spaces[0],mesh,material,formulation,fem_solver,
                    Eulerx, TotalDisp, NeumannForces, M, velocities, accelerations, Increment)
                formulation.power_dissipation.append(power_info[0])
                formulation.internal_power.append(power_info[1])
                formulation.kinetic_power.append(power_info[2])
                formulation.external_power.append(power_info[3])


            # PRINT LOG IF ASKED FOR
            if fem_solver.print_incremental_log:
                dmesh = Mesh()
                dmesh.points = TotalDisp[:,:formulation.ndim,Increment]
                dmesh_bounds = dmesh.Bounds
                if formulation.fields == "electro_mechanics":
                    _bounds = np.zeros((2,formulation.nvar))
                    _bounds[:,:formulation.ndim] = dmesh_bounds
                    _bounds[:,-1] = [TotalDisp[:,-1,Increment].min(),TotalDisp[:,-1,Increment].max()]
                    print("\nMinimum and maximum incremental solution values at increment {} are \n".format(Increment),_bounds)
                else:
                    print("\nMinimum and maximum incremental solution values at increment {} are \n".format(Increment),dmesh_bounds)

            # SAVE INCREMENTAL SOLUTION IF ASKED FOR
            if fem_solver.save_incremental_solution:
                from scipy.io import savemat
                if fem_solver.incremental_solution_filename is not None:
                    savemat(fem_solver.incremental_solution_filename+"_"+str(Increment),{'solution':TotalDisp[:,:,Increment]},do_compression=True)
                else:
                    raise ValueError("No file name provided to save incremental solution")


            print('\nFinished Load increment', Increment, 'in', time()-t_increment, 'seconds')

            try:
                print('Norm of Residual is',
                    np.abs(la.norm(Residual[boundary_condition.columns_in])/self.NormForces), '\n')
            except RuntimeWarning:
                print("Invalid value encountered in norm of Newton-Raphson residual")

            # STORE THE INFORMATION IF NEWTON-RAPHSON FAILS
            if fem_solver.newton_raphson_failed_to_converge:
                solver.condA = np.NAN
                TotalDisp = TotalDisp[:,:,:Increment-1]
                fem_solver.number_of_load_increments = Increment - 1
                break

            # BREAK AT A SPECIFICED LOAD INCREMENT IF ASKED FOR
            if fem_solver.break_at_increment != -1 and fem_solver.break_at_increment is not None:
                if fem_solver.break_at_increment == Increment:
                    if fem_solver.break_at_increment < LoadIncrement - 1:
                        print("\nStopping at increment {} as specified\n\n".format(Increment))
                        TotalDisp = TotalDisp[:,:,:Increment]
                        fem_solver.number_of_load_increments = Increment
                    break

        return TotalDisp


    def NewtonRaphson(self, function_spaces, formulation, solver,
        Increment, K, D, M, NodalForces, Residual, mesh, Eulerx, Eulerp, material,
        boundary_condition, AppliedDirichletInc, fem_solver, velocities, accelerations):

        Tolerance = fem_solver.newton_raphson_tolerance
        LoadIncrement = fem_solver.number_of_load_increments
        LoadFactor = fem_solver.total_time/fem_solver.number_of_load_increments
        Iter = 0

        # EulerxPrev = np.copy(Eulerx)
        # EulerVPrev = np.copy(velocities[:,:,Increment-1])
        # EulerAPrev = np.copy(accelerations[:,:,Increment-1])

        # PREDICTOR STEP
        tmpV = (1. - self.gamma/self.beta)*velocities + (1. - self.gamma/2./self.beta)*LoadFactor*accelerations
        tmpA = (-1./self.beta/LoadFactor)*velocities - (1./2./self.beta)*(1.- 2.*self.beta)*accelerations
        velocities    = tmpV
        accelerations = tmpA

        if formulation.fields == "electro_mechanics":
            M_mech = M[self.mechanical_dofs,:][:,self.mechanical_dofs]
            InertiaResidual = np.zeros((Residual.shape[0],1))
            InertiaResidual[self.mechanical_dofs,0] = M_mech.dot(accelerations.ravel())
            if fem_solver.include_physical_damping:
                D_mech = D[self.mechanical_dofs,:][:,self.mechanical_dofs]
                InertiaResidual[self.mechanical_dofs,0] += D_mech.dot(velocities.ravel())
        else:
            InertiaResidual = np.zeros((Residual.shape[0],1))
            InertiaResidual[:,0] = M.dot(accelerations.ravel())
            if fem_solver.include_physical_damping:
                InertiaResidual[:,0] += D.dot(velocities.ravel())
        Residual[boundary_condition.columns_in] += InertiaResidual[boundary_condition.columns_in]


        # APPLY INCREMENTAL DIRICHLET PER LOAD STEP (THIS IS INCREMENTAL NOT ACCUMULATIVE)
        IncDirichlet = boundary_condition.UpdateFixDoFs(AppliedDirichletInc,
            K.shape[0],formulation.nvar)
        # UPDATE EULERIAN COORDINATE
        Eulerx += IncDirichlet[:,:formulation.ndim]
        Eulerp[:] = IncDirichlet[:,-1] # ENSURES Eulerp IS CONTIGUOUS - NECESSARY FOR LOW-LEVEL DISPATCHER


        while np.abs(la.norm(Residual[boundary_condition.columns_in])/self.NormForces) > Tolerance or Iter==0:

            # GET EFFECTIVE STIFFNESS
            # K += (1./self.beta/LoadFactor**2)*M
            K += (self.gamma/self.beta/LoadFactor)*D + (1./self.beta/LoadFactor**2)*M
            # GET THE REDUCED SYSTEM OF EQUATIONS
            K_b, F_b, _ = boundary_condition.GetReducedMatrices(K,Residual)

            # SOLVE THE SYSTEM
            sol = solver.Solve(K_b,-F_b)

            # GET ITERATIVE SOLUTION
            dU = boundary_condition.UpdateFreeDoFs(sol,K.shape[0],formulation.nvar)

            # UPDATE THE EULERIAN COMPONENTS
            # UPDATE THE GEOMETRY
            Eulerx += dU[:,:formulation.ndim]
            # GET ITERATIVE ELECTRIC POTENTIAL
            Eulerp += dU[:,-1]

            # UPDATE VELOCITY AND ACCELERATION
            velocities    += self.gamma/self.beta/LoadFactor*dU[:,:formulation.ndim]
            accelerations += 1./self.beta/LoadFactor**2*dU[:,:formulation.ndim]

            # OR ALTERNATIVELY
            # dumA = 1./self.beta/LoadFactor**2*(Eulerx - EulerxPrev) -\
            #     1./self.beta/LoadFactor*(EulerVPrev) -\
            #     1./2./self.beta*(1. - 2.*self.beta)*(EulerAPrev)
            # dumV = (1. - self.gamma/self.beta)*(EulerVPrev) +\
            #     (1. - self.gamma/2./self.beta)*LoadFactor*(EulerAPrev) +\
            #     self.gamma/self.beta/LoadFactor*(Eulerx - EulerxPrev)
            # velocities    = dumV
            # accelerations = dumA

            # RE-ASSEMBLE - COMPUTE STIFFNESS AND INTERNAL TRACTION FORCES
            K, TractionForces, _, _ = Assemble(fem_solver,function_spaces[0], formulation, mesh, material,
                Eulerx, Eulerp)

            # FIND INITIAL RESIDUAL
            if formulation.fields == "electro_mechanics":
                InertiaResidual = np.zeros((TractionForces.shape[0],1))
                InertiaResidual[self.mechanical_dofs,0] = M_mech.dot(accelerations.ravel())
                if fem_solver.include_physical_damping:
                    InertiaResidual[self.mechanical_dofs,0] += D_mech.dot(velocities.ravel())
            else:
                InertiaResidual = np.zeros((TractionForces.shape[0],1))
                InertiaResidual[:,0] = M.dot(accelerations.ravel())
                if fem_solver.include_physical_damping:
                    InertiaResidual[:,0] += D.dot(velocities.ravel())


            # UPDATE RESIDUAL
            Residual[boundary_condition.columns_in] = TractionForces[boundary_condition.columns_in] \
            - NodalForces[boundary_condition.columns_in] + InertiaResidual[boundary_condition.columns_in]


            # SAVE THE NORM
            self.rel_norm_residual = la.norm(Residual[boundary_condition.columns_in])
            if Iter==0:
                self.NormForces = la.norm(Residual[boundary_condition.columns_in])
            self.norm_residual = np.abs(la.norm(Residual[boundary_condition.columns_in])/self.NormForces)

            # SAVE THE NORM
            self.NRConvergence['Increment_'+str(Increment)] = np.append(self.NRConvergence['Increment_'+str(Increment)],\
                self.norm_residual)

            print("Iteration {} for increment {}.".format(Iter, Increment) +\
                " Residual (abs) {0:>16.7g}".format(self.rel_norm_residual),
                "\t Residual (rel) {0:>16.7g}".format(self.norm_residual))

            # BREAK BASED ON RELATIVE NORM
            if np.abs(self.rel_norm_residual) < Tolerance:
                break

            # BREAK BASED ON INCREMENTAL SOLUTION - KEEP IT AFTER UPDATE
            if norm(dU) <=  fem_solver.newton_raphson_solution_tolerance:
                print("Incremental solution within tolerance i.e. norm(dU): {}".format(norm(dU)))
                break

            # UPDATE ITERATION NUMBER
            Iter +=1

            if Iter==fem_solver.maximum_iteration_for_newton_raphson and formulation.fields == "electro_mechanics":
                raise StopIteration("\n\nNewton Raphson did not converge! Maximum number of iterations reached.")

            if Iter==fem_solver.maximum_iteration_for_newton_raphson:
                fem_solver.newton_raphson_failed_to_converge = True
                break

            if np.isnan(self.norm_residual) or self.norm_residual>1e06:
                fem_solver.newton_raphson_failed_to_converge = True
                break

            # USER DEFINED CRITERIA TO BREAK OUT OF NEWTON-RAPHSON
            if fem_solver.user_defined_break_func != None:
                if fem_solver.user_defined_break_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
                    break

            # USER DEFINED CRITERIA TO STOP NEWTON-RAPHSON AND THE WHOLE ANALYSIS
            if fem_solver.user_defined_stop_func != None:
                if fem_solver.user_defined_stop_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
                    fem_solver.newton_raphson_failed_to_converge = True
                    break

        return Eulerx, Eulerp, K, Residual, velocities, accelerations












#------------------------------------------ LINEAR IMPLICIT SOLVER ----------------------------------------------#
#----------------------------------------------------------------------------------------------------------------#
#----------------------------------------------------------------------------------------------------------------#


class LinearImplicitStructuralDynamicIntegrator(StructuralDynamicIntegrator):
    """Implicit dynamic solver for linear problems based on Newmark's beta
    """

    def __init__(self,**kwargs):
        super(LinearImplicitStructuralDynamicIntegrator, self).__init__()
        self.lump_rhs = False
        self.gamma   = 0.5
        self.beta    = 0.25


    def Solver(self, function_spaces, formulation, solver,
        K, M, NeumannForces, NodalForces, Residual,
        mesh, TotalDisp, Eulerx, Eulerp, material, boundary_condition, fem_solver):

        # CHECK FORMULATION
        if formulation.fields != "mechanics" and formulation.fields != "electro_mechanics":
            raise NotImplementedError("Linear implicit solver for {} is not available".format(formulation.fields))

        # GET BOUNDARY CONDITIONS INFROMATION
        self.GetBoundaryInfo(mesh, formulation, boundary_condition)

        LoadIncrement = fem_solver.number_of_load_increments
        LoadFactor = fem_solver.total_time/LoadIncrement

        post_process = PostProcess(formulation.ndim,formulation.nvar)
        post_process.SetAnalysis(analysis_type=fem_solver.analysis_type, analysis_nature=fem_solver.analysis_nature)

        if NeumannForces.ndim == 2 and NeumannForces.shape[1]==1:
            tmp = np.zeros((NeumannForces.shape[0],LoadIncrement))
            tmp[:,0] = NeumannForces[:,0]
            NeumannForces = tmp

        # dU = np.zeros((mesh.points.shape[0]*formulation.nvar))
        # dU[boundary_condition.columns_out] = boundary_condition.applied_dirichlet[:,0]
        # TotalDisp[:,:formulation.nvar,0] = dU.reshape(mesh.points.shape[0],formulation.nvar)
        dU = boundary_condition.UpdateFixDoFs(boundary_condition.applied_dirichlet[:,0], mesh.points.shape[0]*formulation.nvar, formulation.nvar)
        TotalDisp[:,:formulation.nvar,0] = dU
        # INITIALISE VELOCITY AND ACCELERATION
        velocities     = np.zeros((mesh.points.shape[0]*formulation.ndim))
        accelerations  = np.zeros((mesh.points.shape[0]*formulation.ndim))
        # COMPUTE DAMPING MATRIX BASED ON MASS
        D = 0.0
        if fem_solver.include_physical_damping:
            D = fem_solver.damping_factor*M

        if formulation.fields == "electro_mechanics":
            M_mech = M[self.mechanical_dofs,:][:,self.mechanical_dofs]
            if self.include_physical_damping:
                D_mech = D[self.mechanical_dofs,:][:,self.mechanical_dofs]
        else:
            M_mech = M
            D_mech = D

        # COMPUTE INITIAL ACCELERATION FOR TIME STEP 0
        Residual = np.zeros_like(Residual)
        InitResidual = Residual + NeumannForces[:,0][:,None]
        if formulation.fields == "electro_mechanics":
            accelerations[:] = solver.Solve(M_mech, -InitResidual[self.mechanical_dofs].ravel())
        else:
            accelerations[:] = solver.Solve(M, InitResidual.ravel() )

        # COMPUTE AUGMENTED K (INCLUDES INERTIA EFFECT)
        K          += (self.gamma/self.beta/LoadFactor)*D + (1./self.beta/LoadFactor**2)*M
        # GET REDUCED VARIABLES
        K_b, F_b, _ = boundary_condition.GetReducedMatrices(K,Residual)

        if self.lump_rhs:
            M_mech = M_mech.sum(axis=1).A.ravel() # FOR CSR
            # M_mech = M_mech.sum(axis=0).ravel() # FOR CSC
            if self.include_physical_damping:
                D_mech = D_mech.sum(axis=1).A.ravel()


        for Increment in range(1,LoadIncrement):
            t_increment=time()

            # FIXED INCREMENTAL DIRICHLET
            AppliedDirichletInc = boundary_condition.applied_dirichlet[:,Increment-1]

            # APPLY NEUMANN BOUNDARY CONDITIONS
            DeltaF = NeumannForces[:,Increment][:,None]
            NodalForces = DeltaF

            # ACCUMULATED FORCE
            if fem_solver.include_physical_damping:
                if self.lump_rhs:
                    Residual[self.mechanical_dofs,0] = (1./self.beta/LoadFactor**2)*M_mech*TotalDisp[:,:formulation.ndim,Increment-1].ravel() +\
                        (1./self.beta/LoadFactor)*M_mech*velocities + (0.5/self.beta - 1.)*M_mech*accelerations +\
                        (self.gamma/self.beta/LoadFactor)*D_mech*TotalDisp[:,:formulation.ndim,Increment-1].ravel() +\
                        (self.gamma/self.beta - 1.)*D_mech*velocities -\
                        LoadFactor*((1-self.gamma)-self.gamma*(0.5/self.beta - 1.))*D_mech*accelerations
                else:
                    Residual[self.mechanical_dofs,0] = (1./self.beta/LoadFactor**2)*M_mech.dot(TotalDisp[:,:formulation.ndim,Increment-1].ravel()) +\
                        (1./self.beta/LoadFactor)*M_mech.dot(velocities) + (0.5/self.beta - 1.)*M_mech.dot(accelerations) +\
                        (self.gamma/self.beta/LoadFactor)*D_mech.dot(TotalDisp[:,:formulation.ndim,Increment-1].ravel()) +\
                        (self.gamma/self.beta - 1.)*D_mech.dot(velocities) -\
                        LoadFactor*((1-self.gamma)-self.gamma*(0.5/self.beta - 1.))*D_mech.dot(accelerations)
            else:
                if self.lump_rhs:
                    Residual[self.mechanical_dofs,0] = (1./self.beta/LoadFactor**2)*M_mech*TotalDisp[:,:formulation.ndim,Increment-1].ravel() +\
                        (1./self.beta/LoadFactor)*M_mech*velocities + (0.5/self.beta - 1.)*M_mech*accelerations
                else:
                    Residual[self.mechanical_dofs,0] = (1./self.beta/LoadFactor**2)*M_mech.dot(TotalDisp[:,:formulation.ndim,Increment-1].ravel()) +\
                        (1./self.beta/LoadFactor)*M_mech.dot(velocities) + (0.5/self.beta - 1.)*M_mech.dot(accelerations)
            Residual += DeltaF


            # CHECK CONTACT AND ASSEMBLE IF DETECTED
            if fem_solver.has_contact:
                Eulerx = mesh.points + TotalDisp[:,:formulation.ndim,Increment-1]
                TractionForcesContact = np.zeros_like(Residual)
                TractionForcesContact = fem_solver.contact_formulation.AssembleTractions(mesh,material,Eulerx).ravel()*LoadFactor

                if formulation.fields == "electro_mechanics" or formulation.fields == "flexoelectric":
                    Residual[self.mechanical_dofs,0] -= TractionForcesContact
                elif formulation.fields == "mechanics" or formulation.fields == "couple_stress":
                    Residual[:,0] -= TractionForcesContact
                else:
                    raise NotImplementedError("Contact algorithm for {} is not available".format(formulation.fields))

            # REDUCED ACCUMULATED FORCE
            F_b = boundary_condition.ApplyDirichletGetReducedMatrices(K,Residual,
                boundary_condition.applied_dirichlet[:,Increment],LoadFactor=1.0,
                mass=M,only_residual=True)[boundary_condition.columns_in,0]

            # SOLVE THE SYSTEM
            sol = solver.Solve(K_b, F_b, reuse_factorisation=True)

            dU = post_process.TotalComponentSol(sol, boundary_condition.columns_in,
                boundary_condition.columns_out, AppliedDirichletInc,0,K.shape[0])

            # STORE TOTAL SOLUTION DATA
            TotalDisp[:,:,Increment] += dU

            # UPDATE VELOCITY AND ACCELERATION
            accelerations_old = np.copy(accelerations)
            accelerations = (1./self.beta/LoadFactor**2)*(TotalDisp[:,:formulation.ndim,Increment] -\
                TotalDisp[:,:formulation.ndim,Increment-1]).ravel() -\
                1./self.beta/LoadFactor*velocities + (1.-0.5/self.beta)*accelerations_old
            velocities += LoadFactor*(self.gamma*accelerations + (1-self.gamma)*accelerations_old)

            # UPDATE
            Eulerx += dU[:,:formulation.ndim]
            Eulerp += dU[:,-1]

            # LOG REQUESTS
            fem_solver.LogSave(formulation, TotalDisp, Increment)

            # BREAK AT A SPECIFICED LOAD INCREMENT IF ASKED FOR
            if fem_solver.break_at_increment != -1 and fem_solver.break_at_increment is not None:
                if fem_solver.break_at_increment == Increment:
                    if fem_solver.break_at_increment < LoadIncrement - 1:
                        print("\nStopping at increment {} as specified\n\n".format(Increment))
                        TotalDisp = TotalDisp[:,:,:Increment]
                        fem_solver.number_of_load_increments = Increment
                    break

            # STORE THE INFORMATION IF THE SOLVER BLOWS UP
            if Increment > 0:
                U0 = TotalDisp[:,:,Increment-1].ravel()
                U = TotalDisp[:,:,Increment].ravel()
                tol = 1e200 if Increment < 5 else 10.
                if np.isnan(norm(U)) or np.abs(U.max()/(U0.max()+1e-14)) > tol:
                    print("Solver blew up! Norm of incremental solution is too large")
                    TotalDisp = TotalDisp[:,:,:Increment]
                    fem_solver.number_of_load_increments = Increment
                    break

            print('Finished Load increment', Increment, 'in', time()-t_increment, 'seconds\n')

        solver.CleanUp()

        return TotalDisp




from __future__ import print_function
import gc, os, sys
import numpy as np
import scipy as sp
import numpy.linalg as la
import scipy.linalg as sla
from numpy.linalg import norm
from time import time
from copy import deepcopy
from warnings import warn
from time import time

from Florence.FiniteElements.Assembly import Assemble, AssembleExplicit
from Florence import Mesh
from Florence.PostProcessing import PostProcess
from .StructuralDynamicIntegrator import StructuralDynamicIntegrator

__all__ = ["ExplicitStructuralDynamicIntegrator"]


class ExplicitStructuralDynamicIntegrator(StructuralDynamicIntegrator):
    """Generic explicit structural time integerator based on central difference"""

    def __init__(self):
        super(ExplicitStructuralDynamicIntegrator, self).__init__()


    def Solver(self, function_spaces, formulation, solver,
        TractionForces, M, NeumannForces, NodalForces, Residual,
        mesh, TotalDisp, Eulerx, Eulerp, material, boundary_condition, fem_solver):

        # CHECK FORMULATION
        if formulation.fields != "mechanics" and formulation.fields != "electro_mechanics":
            raise NotImplementedError("Explicit solver for {} is not available".format(formulation.fields))

        # GET BOUNDARY CONDITIONS INFROMATION
        self.GetBoundaryInfo(mesh, formulation, boundary_condition)

        # COMPUTE INVERSE OF LUMPED MASS MATRIX
        if formulation.fields == "electro_mechanics":
            if fem_solver.mass_type == "lumped":
                M = M.ravel()
                invM = np.zeros_like(M)
                invM[self.mechanical_dofs] = np.reciprocal(M[self.mechanical_dofs])
                # M_mech = M[self.mechanical_dofs]
                M_mech = M[self.mechanical_dofs]
            else:
                M_mech = M[self.mechanical_dofs,:][:,self.mechanical_dofs]
        else:
            if fem_solver.mass_type == "lumped":
                M = M.ravel()
                M_mech = M
                invM = np.reciprocal(M)
            else:
                M_mech = M

        # COMPUTE DAMPING MATRIX BASED ON MASS
        if fem_solver.include_physical_damping:
            raise NotImplementedError("Damping is not included in the explicit solver")

        TractionForces = TractionForces.ravel()
        Residual = Residual.ravel()

        LoadIncrement = fem_solver.number_of_load_increments
        LoadFactor = fem_solver.total_time/LoadIncrement
        dt = LoadFactor
        AppliedDirichletInc = np.zeros(boundary_condition.applied_dirichlet.shape[0],dtype=np.float64)
        nnode = mesh.points.shape[0]

        if NeumannForces.ndim == 1:
            NeumannForces = NeumannForces[:,None]

        # INITIALISE VELOCITY AND ACCELERATION
        U0      = np.zeros((mesh.points.shape[0]*formulation.ndim))
        V0      = np.zeros((mesh.points.shape[0]*formulation.ndim))
        A0      = np.zeros((mesh.points.shape[0]*formulation.ndim))

        # COMPUTE INITIAL ACCELERATION FOR TIME STEP 0
        InitResidual = (NeumannForces[:,0] - TractionForces).ravel()
        if fem_solver.mass_type == "lumped":
            # A0[:] = (NeumannForces[:,0] - TractionForces).ravel()*invM
            A0[:] = InitResidual[self.mechanical_dofs]*invM[self.mechanical_dofs]
        else:
            #
            if formulation.fields == "electro_mechanics":
                M_b, F_b = M_mech[self.mech_in,:][:,self.mech_in], InitResidual[:,None][self.mech_in,0]
                sola = solver.Solve(M_b,F_b)
                A0[self.mech_in] = sola
            else:
                M_b, F_b = boundary_condition.GetReducedMatrices(M,InitResidual[:,None])[:2]
                sola = solver.Solve(M_b,F_b)
                A0[:] = boundary_condition.UpdateFreeDoFs(sola,TractionForces.shape[0],formulation.ndim).ravel()

        U00  = U0 - dt*V0 + (dt**2/2.)*A0
        U0   = self.UpdateFreeMechanicalDoFs(U0[self.mech_in],formulation.ndim*nnode,formulation.ndim)
        U00  = self.UpdateFreeMechanicalDoFs(U00[self.mech_in],formulation.ndim*nnode,formulation.ndim)

        TotalDisp[:,:formulation.ndim,0] = U00
        TotalDisp[:,:formulation.ndim,1] = U0

        # SET UP THE ELECTROSTATICS SOLVER PARAMETERS ONCE
        if formulation.fields == "electro_mechanics":
            self.SetupElectrostaticsImplicit(mesh, formulation, boundary_condition, material, fem_solver, solver, Eulerx, 0)

        save_counter = 2 if fem_solver.save_frequency == 1 else 1
        # TIME LOOP
        for Increment in range(2,LoadIncrement):

            t_increment = time()

            if boundary_condition.applied_dirichlet.ndim == 2:
                AppliedDirichletInc = boundary_condition.applied_dirichlet[:,Increment-1]
            else:
                # RAMP TYPE LOAD
                AppliedDirichletInc = boundary_condition.applied_dirichlet*(1.*Increment/LoadIncrement)

            # APPLY NEUMANN BOUNDARY CONDITIONS
            if NeumannForces.ndim == 2 and NeumannForces.shape[1]>1:
                NodalForces = NeumannForces[:,Increment-1]
            else:
                # RAMP TYPE LOAD
                NodalForces = NeumannForces.ravel()*(1.*Increment/LoadIncrement)

            # Residual[:] = NodalForces - TractionForces
            Residual = NodalForces - TractionForces
            Residual = Residual[self.mechanical_dofs]

            if fem_solver.mass_type == "lumped":
                # Residual   += (2./dt**2)*M*TotalDisp[:,:,Increment-1].ravel() - (1./dt**2)*M*TotalDisp[:,:,Increment-2].ravel()
                Residual   += (2./dt**2)*M_mech*U0.ravel() - (1./dt**2)*M_mech*U00.ravel()
                U = dt**2*invM[self.mechanical_dofs]*Residual
                U = self.UpdateFreeMechanicalDoFs(U[self.mech_in],formulation.ndim*nnode,formulation.ndim)
                IncDirichlet = self.UpdateFixMechanicalDoFs(AppliedDirichletInc[self.columns_out_mech_reverse_idx],
                    formulation.ndim*nnode,formulation.ndim)

            elif fem_solver.mass_type == "consistent":
                # Residual   += (2./dt**2)*M.dot(U0.ravel()) - (1./dt**2)*M.dot(U00.ravel())
                # F_b = boundary_condition.GetReducedVectors(Residual[:,None],only_residual=True)[0]
                Residual   += (2./dt**2)*M_mech.dot(U0.ravel()) - (1./dt**2)*M_mech.dot(U00.ravel())
                F_b = Residual[:,None][self.mech_in,0]
                U = solver.Solve(M_b,F_b*dt**2)
                U = self.UpdateFreeMechanicalDoFs(U,formulation.ndim*nnode,formulation.ndim)
                IncDirichlet = self.UpdateFixMechanicalDoFs(AppliedDirichletInc[self.columns_out_mech_reverse_idx],
                    formulation.ndim*nnode,formulation.ndim)

            # COMPUTE VELOCITY AND ACCELERATION
            V0[:] = (1./2./dt)*(U+U0).ravel()
            A0[:] = (1./dt**2)*(U-2.*U0+U00).ravel()

            # UPDATE GEOMETRY
            Eulerx[:,:] = mesh.points + U[:,:formulation.ndim]
            Eulerx[:,:] += IncDirichlet[:,:formulation.ndim]

            # SOLVE ELECTROSTATICS PROBLEM
            if formulation.fields == "electro_mechanics":
                Eulerp[:] = self.SolveElectrostaticsImplicit(mesh, formulation, boundary_condition, material, fem_solver, solver, Eulerx, Increment)

            # SAVE RESULTS
            if Increment % fem_solver.save_frequency == 0 or\
                (Increment == LoadIncrement - 1 and save_counter<TotalDisp.shape[2]):
                TotalDisp[:,:formulation.ndim,save_counter] = Eulerx - mesh.points
                if formulation.fields == "electro_mechanics":
                    TotalDisp[:,-1,save_counter] = Eulerp.ravel()
                save_counter += 1

            # STORE THE INFORMATION IF EXPLICIT BLOWS UP
            tol = 1e200 if Increment < 5 else 10.
            if np.isnan(norm(U)) or np.abs(U.max()/(U0.max()+1e-14)) > tol:
                print("Explicit solver blew up! Norm of incremental solution is too large")
                TotalDisp = TotalDisp[:,:,:Increment]
                fem_solver.number_of_load_increments = Increment
                break

            # UPDATE RESULTS FOR NEXT STEP
            U00[:,:formulation.ndim] = U0
            U0[:,:formulation.ndim]  = U

            # ASSEMBLE INTERNAL TRACTION FORCES
            t_assembly = time()
            TractionForces = AssembleExplicit(fem_solver,function_spaces[0], formulation, mesh, material,
                Eulerx, Eulerp)[0].ravel()
            # CHECK CONTACT AND ASSEMBLE IF DETECTED
            if fem_solver.has_contact:
                TractionForcesContact = np.zeros_like(TractionForces)
                TractionForcesContact = fem_solver.contact_formulation.AssembleTractions(mesh,material,Eulerx).ravel()
                if formulation.fields == "electro_mechanics":
                    TractionForces[self.mechanical_dofs] += TractionForcesContact
                elif formulation.fields == "mechanics":
                    TractionForces += TractionForcesContact
                else:
                    raise NotImplementedError("Contact algorithm for {} is not available".format(formulation.fields))

            print("Explicit assembly time is {} seconds".format(time()-t_assembly))


            # COMPUTE DISSIPATION OF ENERGY THROUGH TIME
            if fem_solver.compute_energy_dissipation:
                energy_info = self.ComputeEnergyDissipation(function_spaces[0],mesh,material,formulation,fem_solver,
                    Eulerx, U, NeumannForces, M, V0, Increment)
                formulation.energy_dissipation.append(energy_info[0])
                formulation.internal_energy.append(energy_info[1])
                formulation.kinetic_energy.append(energy_info[2])
                formulation.external_energy.append(energy_info[3])
            # COMPUTE DISSIPATION OF LINEAR MOMENTUM THROUGH TIME
            if fem_solver.compute_linear_momentum_dissipation:
                power_info = self.ComputePowerDissipation(function_spaces[0],mesh,material,formulation,fem_solver,
                    Eulerx, U, NeumannForces, M, V0, A0, Increment)
                formulation.power_dissipation.append(power_info[0])
                formulation.internal_power.append(power_info[1])
                formulation.kinetic_power.append(power_info[2])
                formulation.external_power.append(power_info[3])


            # LOG IF ASKED FOR
            self.LogSave(fem_solver, formulation, U, Eulerp, Increment)

            print('\nFinished Load increment', Increment, 'in', time()-t_increment, 'seconds')

            # BREAK AT A SPECIFICED LOAD INCREMENT IF ASKED FOR
            if fem_solver.break_at_increment != -1 and fem_solver.break_at_increment is not None:
                if fem_solver.break_at_increment == Increment:
                    if fem_solver.break_at_increment < LoadIncrement - 1:
                        print("\nStopping at increment {} as specified\n\n".format(Increment))
                        TotalDisp = TotalDisp[:,:,:Increment]
                        fem_solver.number_of_load_increments = Increment
                    break

        if fem_solver.save_frequency != 1:
            if TotalDisp.shape[2] > save_counter:
                # IN CASE EXPLICIT SOLVER BLEW UP
                TotalDisp = TotalDisp[:,:,:save_counter]
                fem_solver.number_of_load_increments = TotalDisp.shape[2]
            else:
                fem_solver.number_of_load_increments = save_counter

        return TotalDisp


    def SetupElectrostaticsImplicit(self, mesh, formulation, boundary_condition, material, fem_solver, solver, Eulerx, Increment):
        """setup implicit electrostatic problem
        """

        from Florence import BoundaryCondition, FEMSolver, LaplacianSolver, IdealDielectric, AnisotropicIdealDielectric
        from Florence.VariationalPrinciple import LaplacianFormulation, ExplicitPenaltyContactFormulation

        # EMULATE ELECTROSTATICS MODEL
        emesh = deepcopy(mesh)

        if fem_solver.activate_explicit_multigrid:
            # WE GET THE MAX EPS - NOT ELEGANT BUT SEEMINGLY WORKS WELL
            eps_s = []
            for key, value in list(material.__dict__.items()):
                if "eps" in key:
                    eps_s.append(value)
            max_eps = max(eps_s)
            ematerial = IdealDielectric(emesh.InferSpatialDimension(),eps_1=max_eps)

            eanalysis_nature = "linear"
            eoptimise = True
        else:
            ematerial = deepcopy(material)
            ematerial.Hessian = ematerial.Permittivity
            ematerial.KineticMeasures = ematerial.ElectrostaticMeasures
            ematerial.H_VoigtSize = formulation.ndim
            ematerial.nvar = 1
            ematerial.fields = "electrostatics"

            eanalysis_nature = "nonlinear"
            eoptimise = False

        # SET UP BOUNDARY CONDITION DURING SOLUTION
        eboundary_condition = BoundaryCondition()

        eformulation = LaplacianFormulation(mesh)
        efem_solver = FEMSolver(number_of_load_increments=1,analysis_nature=eanalysis_nature,
            newton_raphson_tolerance=fem_solver.newton_raphson_tolerance,optimise=eoptimise)

        self.emesh = emesh
        self.ematerial = ematerial
        self.eformulation = eformulation
        self.eboundary_condition = eboundary_condition
        self.efem_solver = efem_solver


    def SolveElectrostaticsImplicit(self, mesh, formulation, boundary_condition, material, fem_solver, solver, Eulerx, Increment):
        """Solve implicit electrostatic problem
        """

        LoadIncrement = fem_solver.number_of_load_increments
        # IF ALL ELECTRIC DoFs ARE FIXED
        if mesh.points.shape[0] == self.electric_out.shape[0]:
            if self.applied_dirichlet_electric.ndim == 2:
                return self.applied_dirichlet_electric[:,Increment]
            else:
                # RAMP TYPE
                return self.applied_dirichlet_electric*(1.*Increment/LoadIncrement)

        # GET BOUNDARY CONDITIONS
        if boundary_condition.dirichlet_flags.ndim==3:
            self.eboundary_condition.dirichlet_flags = boundary_condition.dirichlet_flags[:,-1,Increment]
        else:
            # RAMP TYPE
            self.eboundary_condition.dirichlet_flags = boundary_condition.dirichlet_flags[:,-1]*(1.*Increment/LoadIncrement)
        if boundary_condition.neumann_flags is not None:
            if boundary_condition.neumann_data_applied_at == "node":
                if boundary_condition.neumann_flags.ndim==3:
                    self.eboundary_condition.neumann_flags = boundary_condition.neumann_flags[:,-1, Increment]
                else:
                    # RAMP TYPE
                    self.eboundary_condition.neumann_flags = boundary_condition.neumann_flags[:,-1]*(1.*Increment/LoadIncrement)
            if boundary_condition.neumann_data_applied_at == "face":
                if boundary_condition.neumann_flags.ndim==2:
                    self.eboundary_condition.neumann_flags = boundary_condition.neumann_flags[:,Increment]
                    self.eboundary_condition.applied_neumann = boundary_condition.applied_neumann[:,-1, Increment]
                    self.eboundary_condition.applied_neumann = self.eboundary_condition.applied_neumann[:,None]
                else:
                    # RAMP TYPE
                    self.eboundary_condition.neumann_flags = boundary_condition.neumann_flags[:]
                    self.eboundary_condition.applied_neumann = boundary_condition.applied_neumann[:,-1]*(1.*Increment/LoadIncrement)
                    self.eboundary_condition.applied_neumann = self.eboundary_condition.applied_neumann[:,None]

        # FILTER OUT CASES WHERE BOUNDARY CONDITION IS APPLIED BUT IS ZERO - RELEASE LOAD CYCLE IN DYNAMICS
        if np.allclose(self.eboundary_condition.dirichlet_flags[~np.isnan(self.eboundary_condition.dirichlet_flags)],0.):
            if self.eboundary_condition.neumann_flags is not None:
                if np.allclose(self.eboundary_condition.neumann_flags[~np.isnan(self.eboundary_condition.neumann_flags)],0.):
                    return 0.

        print("\nSolving the electrostatics problem iteratively")
        esolution = self.efem_solver.Solve(formulation=self.eformulation, mesh=self.emesh,
            material=self.ematerial, boundary_condition=self.eboundary_condition, solver=solver, Eulerx=Eulerx)
        print("Finished solving the electrostatics problem\n")

        return esolution.sol.ravel()








from __future__ import print_function
import numpy as np 
from Florence.QuadratureRules.FeketePointsTri import *
from Florence.QuadratureRules.EquallySpacedPoints import EquallySpacedPointsTri
from Florence.FunctionSpace.JacobiPolynomials.JacobiPolynomials import *
from Florence.FunctionSpace.DegenerateMappings import MapXiEta2RS

def hpBases(C, xi, eta, Transform=0, EvalOpt=0, equally_spaced=False):
    """
        Transform:                  transform to from degenrate quad
        EvalOpt:                    evaluate 1 as an approximation 0.9999999
    """

    eps = FeketePointsTri(C)
    if equally_spaced:
        eps = EquallySpacedPointsTri(C)
        
    N = eps.shape[0]
    # Make the Vandermonde matrix
    V = np.zeros((N,N),dtype=np.float64)

    for i in range(0,N):
        x = eps[i,:]
        p1 = NormalisedJacobiTri(C,x)
        V[i,:] = p1



    nsize = int((C+2)*(C+3)/2.)

    Bases = np.zeros((nsize,1))
    gBases = np.zeros((nsize,2))

    # GradNormalisedJacobiTri IS ALWAYS CALLED WITH (QUAD) R,S COORDINATE
    if Transform:
        # IF TRANSFORMATION TO QUAD COORDINATE IS NEEDED
        r, s = MapXiEta2RS(xi,eta)
        p,dp_dxi,dp_deta = GradNormalisedJacobiTri(C,np.array([r,s]),EvalOpt)
    else:
        # IF XI,ETA ARE DIRECTLY GIVEN IN QUAD FORMAT
        p,dp_dxi,dp_deta = GradNormalisedJacobiTri(C,np.array([xi,eta]),EvalOpt)

    Bases = np.linalg.solve(V.T,p)
    gBases[:,0] = np.linalg.solve(V.T,dp_dxi)
    gBases[:,1] = np.linalg.solve(V.T,dp_deta)

    return Bases, gBases



def GradhpBases(C,r,s):
    print('For nodal triangular bases gradient of bases is computed within the hpBases itself')




#cython: profile=False
#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False

import numpy as np
cimport numpy as np

ctypedef double Real

cdef extern from "_TractionDF_.h":

    void _TractionDF_Filler_(Real *traction,
        const Real* SpatialGradient,
        const Real* CauchyStressTensor,
        const Real* detJ,
        int ngauss,
        int noderpelem,
        int ndim,
        int nvar,
        int H_VoigtSize,
        int requires_geometry_update) nogil


def __TractionIntegrandDF__(np.ndarray[Real, ndim=3, mode='c'] SpatialGradient,
    np.ndarray[Real, ndim=3, mode='c'] CauchyStressTensor,
    np.ndarray[Real, ndim=1] detJ,
    int H_VoigtSize,
    int nvar,
    int requires_geometry_update):

    cdef int ngauss = SpatialGradient.shape[0]
    cdef int nodeperelem = SpatialGradient.shape[1]
    cdef int ndim = SpatialGradient.shape[2]
    cdef int local_size = nvar*nodeperelem

    cdef np.ndarray[Real, ndim=2, mode='c'] traction = np.zeros((local_size,1),dtype=np.float64)

    _TractionDF_Filler_(&traction[0,0],
        &SpatialGradient[0,0,0],
        &CauchyStressTensor[0,0,0],
        &detJ[0],
        ngauss,
        nodeperelem,
        ndim,
        nvar,
        H_VoigtSize,
        requires_geometry_update)

    return traction


from .MaterialBase import Material

# LINEAR MATERIAL MODELS
from .LinearElastic import *
from .IncrementalLinearElastic import *
from .TranservselyIsotropicLinearElastic import *
# NONLINEAR MATERIAL MODELS
from .NeoHookean import *
from .NeoHookean_1 import *
from .NeoHookeanCoercive import *
from .NearlyIncompressibleNeoHookean import *
from .MooneyRivlin import *
from .MooneyRivlin_1 import *
from .NearlyIncompressibleMooneyRivlin import *

# REGULARISED MATERIAL MODELS
from .RegularisedNeoHookean import *

# ANISOTROPIC MODELS
from .TranservselyIsotropicHyperElastic import *
from .BonetTranservselyIsotropicHyperElastic import *
from .AnisotropicMooneyRivlin_0 import *
from .AnisotropicMooneyRivlin_1 import *

# ELECTROMECHANICAL MATERIAL MODELS - ENTHALPY
from .IsotropicElectroMechanics_0 import *
from .IsotropicElectroMechanics_1 import *
from .IsotropicElectroMechanics_2 import *
from .IsotropicElectroMechanics_3 import *
from .SteinmannModel import *

# ELECTROMECHANICAL MATERIAL MODELS - INTERNAL ENERGY
from .IsotropicElectroMechanics_100 import *
from .IsotropicElectroMechanics_101 import *
from .IsotropicElectroMechanics_102 import *
from .IsotropicElectroMechanics_103 import *
from .IsotropicElectroMechanics_104 import *
from .IsotropicElectroMechanics_105 import *
from .IsotropicElectroMechanics_106 import *
from .IsotropicElectroMechanics_107 import *
from .IsotropicElectroMechanics_108 import *
from .Piezoelectric_100 import *

from .IsotropicElectroMechanics_109 import *


# HELMHOLTZ AND INTERNAL ENERGY EQUIVALENT MODELS
from .IsotropicElectroMechanics_200 import *
from .IsotropicElectroMechanics_201 import *


# COMPOSTITES
from .Multi_IsotropicElectroMechanics_101 import *
from .Multi_Piezoelectric_100 import *

# EXPLICIT MODELS
from .ExplicitMooneyRivlin import *
from .ExplicitIsotropicElectroMechanics_108 import *


# ELECTROSTATIC MODELS
from .IdealDielectric import *
from .AnisotropicIdealDielectric import *


# COUPLE STRESS MODELS
from .CoupleStressModel import *

# COUPLE STRESS BASED FLEXOELECTRIC MODELS
from .IsotropicLinearFlexoelectricModel import *


import numpy as np
from copy import deepcopy
from warnings import warn
from .Mesh import Mesh
from .GeometricPath import *
from Florence.Tensor import totuple


__all__ = ['HarvesterPatch']

"""
A series of custom meshes
"""


def HarvesterPatch(ndisc=20, nradial=4, show_plot=False):
    """A custom mesh for an energy harvester patch. [Not to be modified]
        ndisc:              [int] number of discretisation in c
        ndradial:           [int] number of discretisation in radial directions for different
                            components of harevester
    """


    center = np.array([30.6979,20.5])
    p1     = np.array([30.,20.])
    p2     = np.array([30.,21.])
    p1line = p1 - center
    p2line = p2 - center
    radius = np.linalg.norm(p1line)
    pp = np.array([center[0],center[1]+radius])
    y_line = pp - center
    start_angle = -np.pi/2. - np.arccos(np.linalg.norm(y_line*p1line)/np.linalg.norm(y_line)/np.linalg.norm(p1line))
    end_angle   = np.pi/2. + np.arccos(np.linalg.norm(y_line*p1line)/np.linalg.norm(y_line)/np.linalg.norm(p1line))
    points = np.array([p1,p2,center])

    # nradial = 4
    mesh = Mesh()
    mesh.Arc(element_type="quad", radius=radius, start_angle=start_angle, 
        end_angle=end_angle, nrad=nradial, ncirc=ndisc, center=(center[0],center[1]), refinement=True)

    mesh1 = Mesh()
    mesh1.Triangle(element_type="quad",npoints=nradial, c1=totuple(center), c2=totuple(p1), c3=totuple(p2))

    mesh += mesh1

    mesh_patch = Mesh()
    mesh_patch.HollowArc(ncirc=ndisc, nrad=nradial, center=(-7.818181,44.22727272), 
        start_angle=np.arctan(44.22727272/-7.818181), end_angle=np.arctan(-24.22727272/37.818181), 
        element_type="quad", inner_radius=43.9129782, outer_radius=44.9129782)

    mesh3 = Mesh()
    mesh3.Triangle(element_type="quad",npoints=nradial, c2=totuple(p1), c3=totuple(p2), c1=(mesh_patch.points[0,0], mesh_patch.points[0,1]))

    mesh += mesh3
    mesh += mesh_patch


    mesh.Extrude(nlong=ndisc,length=40)

    if show_plot:
        mesh.SimplePlot()

    return mesh


def CurvedPlate(ncirc=2, nlong=20, show_plot=False):
    """Custom mesh for plate with curved edges
        ncirc           discretisation around circular fillets
        nlong           discretisation along the length - X
    """

    mesh_arc = Mesh()
    mesh_arc.Arc(element_type="quad",nrad=ncirc,ncirc=ncirc, radius=5)

    mesh_arc1 = deepcopy(mesh_arc)
    mesh_arc1.points[:,1] += 15
    mesh_arc1.points[:,0] += 95
    mesh_arc2 = deepcopy(mesh_arc)
    mesh_arc2.points[:,1] +=15
    mesh_arc2.points[:,0] *= -1.
    mesh_arc2.points[:,0] += 5.

    mesh_plate1 = Mesh()
    mesh_plate1.Rectangle(element_type="quad",lower_left_point=(5,15),upper_right_point=(95,20),ny=ncirc, nx=nlong)

    mesh_plate2 = deepcopy(mesh_plate1)
    mesh_plate2.points[:,1] -= 5.

    mesh_square1 = Mesh()
    mesh_square1.Square(element_type="quad",lower_left_point=(0,10), side_length=5,nx=ncirc,ny=ncirc)

    mesh_square2 = deepcopy(mesh_square1)
    mesh_square2.points[:,0] += 95

    mesh = mesh_plate1 + mesh_plate2 + mesh_arc1 + mesh_arc2 + mesh_square1 + mesh_square2

    mesh.Extrude(length=0.5,nlong=1)

    mesh2 = deepcopy(mesh)
    mesh2.points[:,2] += 0.5
    mesh += mesh2

    if show_plot:
        mesh.SimplePlot()

    return mesh



def Torus(show_plot=False):
    """Custom mesh for torus
    """

    raise NotImplementedError("Not fully implemented yet")

    # MAKE TORUS WORK
    from copy import deepcopy
    from numpy.linalg import norm
    mesh = Mesh()
    mesh.Circle(element_type="quad", ncirc=2, nrad=2)
    tmesh = deepcopy(mesh)
    arc = GeometricArc(start=(10,10,8),end=(10,10,-8))
    # arc.GeometricArc()
    nlong = 10
    points = mesh.Extrude(path=arc, nlong=nlong)
    # mesh.SimplePlot()
    # print points

    # elem_nodes = tmesh.elements[0,:]
    # p1 = tmesh.points[elem_nodes[0],:]
    # p2 = tmesh.points[elem_nodes[1],:]
    # p3 = tmesh.points[elem_nodes[2],:]
    # p4 = tmesh.points[elem_nodes[3],:]
    
    # E1 = np.append(p2 - p1, 0.0)
    # E2 = np.append(p4 - p1, 0.0)
    # E3 = np.array([0,0,1.])

    # E1 /= norm(E1)
    # E2 /= norm(E2)

    # # print E1,E2,E3

    # elem_nodes = mesh.elements[0,:]
    # p1 = mesh.points[elem_nodes[0],:]
    # p2 = mesh.points[elem_nodes[1],:]
    # p3 = mesh.points[elem_nodes[2],:]
    # p4 = mesh.points[elem_nodes[3],:]
    # p5 = mesh.points[elem_nodes[4],:]
    # e1 = p2 - p1
    # e2 = p4 - p1
    # e3 = p5 - p1

    # e1 /= norm(e1)
    # e2 /= norm(e2)
    # e3 /= norm(e3)
    # # print e1,e2,e3


    # # TRANSFORMATION MATRIX
    # Q = np.array([
    #     [np.einsum('i,i',e1,E1), np.einsum('i,i',e1,E2), np.einsum('i,i',e1,E3)],
    #     [np.einsum('i,i',e2,E1), np.einsum('i,i',e2,E2), np.einsum('i,i',e2,E3)],
    #     [np.einsum('i,i',e3,E1), np.einsum('i,i',e3,E2), np.einsum('i,i',e3,E3)]
    #     ])
    # mesh.points = np.dot(mesh.points,Q.T)
    # points = np.dot(points,Q)
    # E1 = np.array([1,0,0.])
    E3 = np.array([0.,0.,1.])
    nnode_2D = tmesh.points.shape[0]
    for i in range(nlong+1):
        # e1 = points[i,:][None,:]/norm(points[i,:])
        # Q = np.dot(E1[:,None],e1)
        # vpoints = np.dot(points,Q)

        e3 = points[i+1,:] - points[i,:]; e3 /= norm(e3)
        Q = np.dot(e3[:,None],E3[None,:])
        # print Q
        # print np.dot(Q,points[i,:][:,None])
        vpoints = np.dot(points,Q)

        # print current_points
        mesh.points[nnode_2D*i:nnode_2D*(i+1),:2] = tmesh.points + points[i,:2]
        mesh.points[nnode_2D*i:nnode_2D*(i+1), 2] = vpoints[i,2]


    # print Q
    # print tmesh.points

    # mesh = Mesh.HexahedralProjection()
    if show_plot:
        mesh.SimplePlot()

    return mesh




import numpy as np

def InteriorPostProcess(total_sol,internal_points,global_coord,element_connectivity,w,z,boundary_elements,C,dN,Basis,Jacobian, nx, ny, XCO, YCO):
    # Computing potential and flux - Interiors

    POT = np.zeros((internal_points.shape[0],1))
    FLUX1 = np.zeros((internal_points.shape[0],1))
    FLUX2 = np.zeros((internal_points.shape[0],1))

    # Loop over collocation points
    for j in range(0,internal_points.shape[0]):
        XP = internal_points[j,0];      YP = internal_points[j,1]
        # Loop over elements
        # for elem in range(0,len(boundary_elements)):
        for elem in range(0,boundary_elements.shape[0]):
            # Loop over nodes of the element
            for i in range(0,C+2):
                # Carry out usual Gaussian integration
                A=0; B=0
                DU1 = 0; DU2=0; DQ1=0; DQ2=0
                for g in range(0,w.shape[0]):
                    # Compute the radial distance
                    RA = np.sqrt((XCO[g,elem]-XP)**2+(YCO[g,elem]-YP)**2)

                    # Compute Kernels - Assuming both sides are multiplied by 2pi
                    K1 = (-1.0/(RA**2))*((XP-XCO[g,elem])*nx[g,elem]+(YP-YCO[g,elem])*ny[g,elem])
                    K2 = np.log(1.0/RA)

                    RD1 = (XCO[g,elem]-XP)/RA
                    RD2 = (YCO[g,elem]-YP)/RA

                    # For potential
                    A+= K1*Basis[i,g]*Jacobian[g,elem]*w[g]
                    B+= K2*Basis[i,g]*Jacobian[g,elem]*w[g]

                    # Derivatives of potential along x and y
                    DU1 +=(1.0/RA**2)*(XCO[g,elem]-XP)*Basis[i,g]*Jacobian[g,elem]*w[g]
                    DU2 +=(1.0/RA**2)*(YCO[g,elem]-YP)*Basis[i,g]*Jacobian[g,elem]*w[g]

                    # Derivatives of flux along x and y
                    DQ1 += -((2.0*(RD1**2)-1.0)*nx[g,elem]+2.0*RD1*RD2*ny[g,elem])*Basis[i,g]*w[g]*Jacobian[g,elem]/(RA**2)
                    DQ2 += -((2.0*(RD2**2)-1.0)*ny[g,elem]+2.0*RD1*RD2*nx[g,elem])*Basis[i,g]*w[g]*Jacobian[g,elem]/(RA**2)

                POT[j] += total_sol[element_connectivity[elem,i],0]*A-total_sol[element_connectivity[elem,i],1]*B
                FLUX1[j] += total_sol[element_connectivity[elem,i],1]*DU1-total_sol[element_connectivity[elem,i],0]*DQ1
                FLUX2[j] += total_sol[element_connectivity[elem,i],1]*DU2-total_sol[element_connectivity[elem,i],0]*DQ2

        # Divide by 2pi
        POT[j] = POT[j]/2.0/np.pi
        FLUX1[j] = FLUX1[j]/2.0/np.pi
        FLUX2[j] = FLUX2[j]/2.0/np.pi


    return POT, FLUX1, FLUX2





def GetTotalSolution(sol,boundary_data,LHS2LHS,RHS2LHS):

    total_sol = np.copy(boundary_data)
    total_sol[np.array(LHS2LHS,dtype=int),0] = sol[np.array(LHS2LHS,dtype=int),0]
    total_sol[np.array(RHS2LHS,dtype=int),1] = sol[np.array(RHS2LHS,dtype=int),0]

    return total_sol


from .NumericIntegrator import GaussQuadrature, GaussQuadratureLogarithmic, GaussLobattoQuadrature
from .GaussLobattoPoints import GaussLobattoPoints1D, GaussLobattoPointsQuad, GaussLobattoPointsHex
from .QuadratureRule import QuadratureRule
from .FeketePointsTri import FeketePointsTri
from .FeketePointsTet import FeketePointsTet
from .GaussLobattoPoints import GaussLobattoPoints1D, GaussLobattoPointsQuad, GaussLobattoPointsHex




class JacobianError(ArithmeticError):
    def __init__(self,value=None):
        self.value = value
    def __str__(self):
        if self.value is None:
            self.value = 'Jacobian of mapping is close to zero'
        return repr(self.value) 


class IllConditionedError(ArithmeticError):
    def __init__(self,value=None):
        self.value = value
    def __str__(self):
        if self.value is None:
            self.value = 'Matrix is ill conditioned'
        return repr(self.value)


import numpy as np
from warnings import warn
from Florence.QuadratureRules import GaussQuadrature, QuadraturePointsWeightsTet, QuadraturePointsWeightsTri


class QuadratureRule(object):

    def __init__(self, qtype="gauss", norder=2, mesh_type="tri", optimal=3, is_flattened=False):
        """
            input:
                is_flattened:           [bool] only used for quads and hexes as tensor based
                                        quadrature is not flattened where tabulated values are
        """

        self.qtype = qtype
        self.norder = norder
        # OPTIMAL QUADRATURE POINTS FOR TRIS AND TETS
        self.optimal = optimal

        if optimal is False or optimal is None:
            self.qtype = None

        z=[]; w=[];

        if mesh_type == "quad" or mesh_type == "hex":
            z, w = GaussQuadrature(self.norder,-1.,1.)
        elif mesh_type == "tet":
            zw = QuadraturePointsWeightsTet.QuadraturePointsWeightsTet(self.norder,optimal)
            z = zw[:,:-1]; z=z.reshape(z.shape[0],z.shape[1]); w=zw[:,-1]
        elif mesh_type == "tri":
            zw = QuadraturePointsWeightsTri.QuadraturePointsWeightsTri(self.norder,optimal)
            z = zw[:,:-1]; z=z.reshape(z.shape[0],z.shape[1]); w=zw[:,-1]
        elif mesh_type == "line":
            z, w = GaussQuadrature(self.norder,-1.,1.)

        self.points = z
        self.weights = w

        if mesh_type == "quad" or mesh_type == "hex":
            if is_flattened:
                self.Flatten(mesh_type=mesh_type)



    def Flatten(self, mesh_type=None):
        """Flateen a quadrature rule given its tensor product form
        """

        if mesh_type == "quad":

            w = np.zeros((int(self.points.shape[0]**2)))
            z = np.zeros((int(self.points.shape[0]**2),2))

            counter = 0
            for i in range(self.points.shape[0]):
                for j in range(self.points.shape[0]):
                    w[counter] = self.weights[i]*self.weights[j]

                    z[counter,0] = self.points[i]
                    z[counter,1] = self.points[j]
                    counter += 1

        elif mesh_type == "hex":

            w = np.zeros((int(self.points.shape[0]**3)))
            z = np.zeros((int(self.points.shape[0]**3),3))

            counter = 0
            for i in range(self.points.shape[0]):
                for j in range(self.points.shape[0]):
                    for k in range(self.points.shape[0]):
                        w[counter] = self.weights[i]*self.weights[j]*self.weights[k]

                        z[counter,0] = self.points[i]
                        z[counter,1] = self.points[j]
                        z[counter,2] = self.points[k]
                        counter += 1

        else:
            raise ValueError("Element type not understood")


        self.points = z
        self.weights = w



from cython import boundscheck, wraparound, cdivision
import numpy as np
cimport numpy as np

from libc.stdint cimport int64_t
from libc.math cimport fabs
from libcpp.vector cimport vector
from cpython cimport bool

__all__ = ['trace','doublecontract','makezero','makezero3d','issymetric','tovoigt', 'tovoigt3', 'cross2d',
'fillin','findfirst','findequal','findequal_approx','findless','findgreater']

# COMPILE USING
# cython --cplus Numeric.pyx
# g++ -std=c++11 -fPIC -shared -pthread -O3 -fwrapv -fno-strict-aliasing -Wall -finline-functions \
# -ffast-math -funroll-loops Numeric.cpp _Numeric.cpp -o Numeric.so -I/usr/include/python2.7 -lpython2.7 -lm


#---------------------------------------------------
# FUSED TYPES ESSENTIALLY INTRODUCE THE SAME OVERHEAD AS THAT OF NUMPY UFUNCS
# ctypedef fused Float:
#     np.float32_t
#     np.float64_t

# ctypedef fused Int:
#     np.int32_t
#     np.int64_t

# ctypedef fused Generic:
#     np.int32_t
#     np.uint32_t
#     np.int64_t
#     np.uint64_t
#     np.float32_t
#     np.float64_t
#---------------------------------------------------


Real = np.float64
ctypedef np.float64_t Real_t

Integer = np.int64
# ctypedef np.int64_t Integer_t
ctypedef int64_t Integer_t

cdef extern from "<algorithm>" namespace "std":
    void fill(Real_t *first, Real_t *last, Real_t num)
    Integer_t* find(Integer_t *first, Integer_t *last, Integer_t num)

cdef extern from "_Numeric.cpp":
    vector[Integer_t] FindEqual(const Integer_t *arr, Integer_t size, Integer_t num)
    vector[Integer_t] FindEqualApprox(const Real_t *arr, Integer_t size,Real_t num, Real_t tolerance)
    vector[Integer_t] FindLessThan(const Real_t *arr, Integer_t size, Real_t num)
    vector[Integer_t] FindGreaterThan(const Real_t *arr, Integer_t size, Real_t num)
    # vector[Generic] FindGreaterThan(const Generic *arr, Integer_t size, Generic num) # NOT ALLOWED



# cpdef ALSO INTRODUCES A BIT OF OVERHEAD - AVOID IT UNLESS NECESSARY

@boundscheck(False)
@wraparound(False)
def trace(np.ndarray[Real_t, ndim=2] A):
    cdef:
        int i
        Real_t trA = 0.0
        int n1 = A.shape[0]
        int n2 = A.shape[1]

    if n1 != n2:
        raise AssertionError("Trace of non-Hermitian (non-square) matrix requested")

    for i in range(0,n1):
       trA += A[i,i]
    return trA



@boundscheck(False)
@wraparound(False)
def doublecontract(np.ndarray[Real_t, ndim=2] A, np.ndarray[Real_t, ndim=2] B):
    """Double contraction of 2D Arrays"""
    # THIS FUNCTION IS FASTER THAN np.einsum FOR SMALLER MATRICES AND SLOWER THAN FOR LARGER MATRICES.
    cdef:
        int a1 = A.shape[0]
        int b1 = B.shape[0]
        int a2 = A.shape[1]
        int b2 = B.shape[1]

    if a1!=b1 or a2!=b2:
        raise ValueError('Incompatible dimensions for double contraction')

    return _doublecontract(&A[0,0],&B[0,0], a1*a2)

# CYTHON DOES NOT ACCEPT THE const KEYWORD FOR A & B HERE FOR SOME REASON
cdef inline Real_t _doublecontract(Real_t *A, Real_t *B, int size):

    cdef int i
    cdef Real_t sum_reducter = 0.
    for i in range(size):
        sum_reducter += A[i]*B[i]
    return sum_reducter


@boundscheck(False)
@wraparound(False)
cpdef void makezero(np.ndarray[Real_t, ndim=2] A, Real_t tol=1.0e-14):
    """Substitute the elements of an array which are close to zero with zero.
        This is an in-place operation and does not return anything"""

    cdef:
        int i,j
        int a1 = A.shape[0]
        int a2 = A.shape[1]

    for i in range(a1):
        for j in range(a2):
            if fabs(A[i,j]) < tol:
                A[i,j] = 0.

@boundscheck(False)
@wraparound(False)
cpdef void makezero3d(np.ndarray[Real_t, ndim=3] A, Real_t tol=1.0e-14):
    """Substitute the elements of an array which are close to zero with zero.
        This is an in-place operation and does not return anything"""

    cdef:
        int i,j,k
        int a1 = A.shape[0]
        int a2 = A.shape[1]
        int a3 = A.shape[2]

    for i in range(a1):
        for j in range(a2):
            for k in range(a3):
                if fabs(A[i,j,k]) < tol:
                    A[i,j,k] = 0.


@boundscheck(False)
@wraparound(False)
cpdef bool issymetric(np.ndarray[Real_t, ndim=2] A, Real_t tol=1.0e-12):
    """Checks if a Hermitian floating point matrix is symmetric within a tolerance"""
    cdef:
        int i,j
        int a1 = A.shape[0]
        int a2 = A.shape[1]
        bool issym = True

    if a1 != a2:
        raise ValueError("Symmetricity of a non-Hermitian (non-square) matrix requested")

    for i in range(a1):
        for j in range(a2):
            if fabs(A[i,j] - A[j,i]) > tol:
                issym = False
                break
        if issym == True:
            break

    return issym


@boundscheck(False)
@wraparound(False)
def tovoigt(np.ndarray[Real_t, ndim=4, mode='c'] C):
    """Convert a 4D array to its Voigt represenation"""
    cdef np.ndarray[Real_t, ndim=2,mode='c'] VoigtA
    cdef int n1dim = C.shape[0]
    # DISPATCH CALL TO APPROPRIATE FUNCTION
    if n1dim == 3:
        VoigtA = np.zeros((6,6),dtype=np.float64)
        _Voigt3(&C[0,0,0,0],&VoigtA[0,0])
    elif n1dim == 2:
        VoigtA = np.zeros((3,3),dtype=np.float64)
        _Voigt2(&C[0,0,0,0],&VoigtA[0,0])

    return VoigtA


cdef _Voigt3(const Real_t *C, Real_t *VoigtA):
    VoigtA[0] = C[0]
    VoigtA[1] = C[4]
    VoigtA[2] = C[8]
    VoigtA[3] = 0.5*(C[1]+C[3])
    VoigtA[4] = 0.5*(C[2]+C[6])
    VoigtA[5] = 0.5*(C[5]+C[7])
    VoigtA[6] = VoigtA[1]
    VoigtA[7] = C[40]
    VoigtA[8] = C[44]
    VoigtA[9] = 0.5*(C[37]+C[39])
    VoigtA[10] = 0.5*(C[38]+C[42])
    VoigtA[11] = 0.5*(C[41]+C[43])
    VoigtA[12] = VoigtA[2]
    VoigtA[13] = VoigtA[8]
    VoigtA[14] = C[80]
    VoigtA[15] = 0.5*(C[73]+C[75])
    VoigtA[16] = 0.5*(C[74]+C[78])
    VoigtA[17] = 0.5*(C[77]+C[79])
    VoigtA[18] = VoigtA[3]
    VoigtA[19] = VoigtA[9]
    VoigtA[20] = VoigtA[15]
    VoigtA[21] = 0.5*(C[10]+C[12])
    VoigtA[22] = 0.5*(C[11]+C[15])
    VoigtA[23] = 0.5*(C[14]+C[16])
    VoigtA[24] = VoigtA[4]
    VoigtA[25] = VoigtA[10]
    VoigtA[26] = VoigtA[16]
    VoigtA[27] = VoigtA[22]
    VoigtA[28] = 0.5*(C[20]+C[24])
    VoigtA[29] = 0.5*(C[23]+C[25])
    VoigtA[30] = VoigtA[5]
    VoigtA[31] = VoigtA[11]
    VoigtA[32] = VoigtA[17]
    VoigtA[33] = VoigtA[23]
    VoigtA[34] = VoigtA[29]
    VoigtA[35] = 0.5*(C[50]+C[52])


cdef _Voigt2(const Real_t *C, Real_t *VoigtA):
    VoigtA[0] = C[0]
    VoigtA[1] = C[3]
    VoigtA[2] = 0.5*(C[1]+C[2])
    VoigtA[3] = VoigtA[1]
    VoigtA[4] = C[15]
    VoigtA[5] = 0.5*(C[13]+C[14])
    VoigtA[6] = VoigtA[2]
    VoigtA[7] = VoigtA[5]
    VoigtA[8] = 0.5*(C[5]+C[6])




@boundscheck(False)
@wraparound(False)
def tovoigt3(np.ndarray[Real_t, ndim=3, mode='c'] e):
    """Convert a 3D array to its Voigt represenation"""
    cdef np.ndarray[Real_t, ndim=2,mode='c'] VoigtA
    cdef int n1dim = e.shape[0]
    # DISPATCH CALL TO APPROPRIATE FUNCTION
    if n1dim == 3:
        VoigtA = np.zeros((6,3),dtype=np.float64)
        _Voigt33(&e[0,0,0],&VoigtA[0,0])
    elif n1dim == 2:
        VoigtA = np.zeros((3,2),dtype=np.float64)
        _Voigt23(&e[0,0,0],&VoigtA[0,0])

    return VoigtA


cdef _Voigt33(const Real_t *e, Real_t *VoigtA):
    VoigtA[0]  = e[0]
    VoigtA[1]  = e[1]
    VoigtA[2]  = e[2]
    VoigtA[3]  = e[12]
    VoigtA[4]  = e[13]
    VoigtA[5]  = e[14]
    VoigtA[6]  = e[24]
    VoigtA[7]  = e[25]
    VoigtA[8]  = e[26]
    VoigtA[9]  = 0.5*(e[3]+e[9])
    VoigtA[10] = 0.5*(e[4]+e[10])
    VoigtA[11] = 0.5*(e[5]+e[11])
    VoigtA[12] = 0.5*(e[6]+e[18])
    VoigtA[13] = 0.5*(e[7]+e[19])
    VoigtA[14] = 0.5*(e[8]+e[20])
    VoigtA[15] = 0.5*(e[15]+e[21])
    VoigtA[16] = 0.5*(e[16]+e[22])
    VoigtA[17] = 0.5*(e[17]+e[23])


cdef _Voigt23(const Real_t *e, Real_t *VoigtA):
    VoigtA[0] = e[0]
    VoigtA[1] = e[1]
    VoigtA[2] = e[6]
    VoigtA[3] = e[7]
    VoigtA[4] = 0.5*(e[2]+e[4])
    VoigtA[5] = 0.5*(e[3]+e[5])


@boundscheck(False)
def cross2d(np.ndarray[double, ndim=2] A, np.ndarray[double, ndim=2] B, str dim="3d", toarray=False):
    """Cross product of second order tensors A_ij x B_ij defined in the sense of R. de Boer
        [Vektor- und Tensorrechnung fur Ingenieure] and J. Bonet [A computational framework
        for polyconvex large strain elasticity].

        input:
            A:                      [ndarray] ndarry of 3x3
            B:                      [ndarray] ndarry of 3x3
            dim:                    [str] either "2d" or "3d". To get a cross product in a 2D
                                    space you still need to supply the matrices in 3D space
                                    and specify the dim argument as 2d. The dim="2d" essentially
                                    assumes that A and B have their zeros in the third direction
                                    as in:
                                        A [and B] = np.array([
                                                            [A00, A01, 0],
                                                            [A10, A11, 0],
                                                            [  0,   0, 0],
                                                            ])

            toarray:                [bool] The default outupt of this unction is a python list
                                    and not a numpy array as one would expect. This is for
                                    performance reason, as numpy for 3x3 array would introduce
                                    quite a bit of overhead. Specify toarray to true if a numpy
                                    array is desired as an output
    """


    if A.shape[0] != 3 or A.shape[1] != 3:
        raise ValueError("Dimension of matrix should be 3x3")

    cdef:
        Real_t A00=A[0,0]
        Real_t A11=A[1,1]
        Real_t A22=A[2,2]
        Real_t A01=A[0,1]
        Real_t A02=A[0,2]
        Real_t A12=A[1,2]
        Real_t A10=A[1,0]
        Real_t A20=A[2,0]
        Real_t A21=A[2,1]

        Real_t B00=B[0,0]
        Real_t B11=B[1,1]
        Real_t B22=B[2,2]
        Real_t B01=B[0,1]
        Real_t B02=B[0,2]
        Real_t B12=B[1,2]
        Real_t B10=B[1,0]
        Real_t B20=B[2,0]
        Real_t B21=B[2,1]

    cdef double AB[3][3]

    if dim == "3d":
        AB[:] = [
            [ A11*B22 - A12*B21 - A21*B12 + A22*B11, A12*B20 - A10*B22 + A20*B12 - A22*B10, A10*B21 - A11*B20 - A20*B11 + A21*B10],
            [ A02*B21 - A01*B22 + A21*B02 - A22*B01, A00*B22 - A02*B20 - A20*B02 + A22*B00, A01*B20 - A00*B21 + A20*B01 - A21*B00],
            [ A01*B12 - A02*B11 - A11*B02 + A12*B01, A02*B10 - A00*B12 + A10*B02 - A12*B00, A00*B11 - A01*B10 - A10*B01 + A11*B00]
            ]

    elif dim == "2d":
        # AB[:] = [
        #     [ 0., 0., 0.],
        #     [ 0., 0., 0.],
        #     [ 0., 0., A00*B11 - A01*B10 - A10*B01 + A11*B00]
        #     ]
        AB[2][2] = A00*B11 - A01*B10 - A10*B01 + A11*B00

    if toarray:
        return np.array(AB,copy=False)
    return AB



@boundscheck(False)
@cdivision(True)
def findfirst(np.ndarray A, Integer_t num):
    """Find the first occurence of a value in an ndarray"""

    if A.dtype != Integer:
        raise TypeError("Expected an array of integers")

    cdef:
        Integer_t size = A.size
        Integer_t ndim = A.ndim
        Integer_t[::1] arr = A.ravel()
        Integer_t col = A.shape[1]

    cdef Integer_t *idx = find(&arr[0],<Integer_t*>(&arr[-1]+1),num)
    cdef Integer_t ret = idx - &arr[0]

    if ret == size:
        return ()

    if ndim == 1:
        return ret
    elif ndim == 2:
        return (ret // col, ret % col)
    else:
        raise ValueError("Dimension mismatch. Array should be either 1D or 2D")


@boundscheck(False)
def fillin(np.ndarray A, Real_t num):
    """Fill ndarray with a value. This is an in-place operation"""

    if A.dtype != Real:
        raise TypeError("Expected an array of doubles")

    cdef Real_t[::1] arr = A.ravel()
    fill(&arr[0],<Real_t*>(&arr[-1]+1),num)
    arr[-1] = num


@boundscheck(False)
@wraparound(False)
def findequal(np.ndarray A, Integer_t num):
    """An equivalent function to numpy.where(A==num). The difference here is that
        numpy.where results in a boolean array of size equal to the original
        array A, even if number of counts is none or a single element. This function
        allocates and does push_back on per count"""

    if A.dtype != Integer:
        raise TypeError("Type mismatch. Array and value must have the same type")

    cdef Integer_t[::1] arr = A.ravel()
    cdef vector[Integer_t] idx = FindEqual(&arr[0],A.size,num)

    if A.ndim == 1:
        return np.array(idx,copy=False)
    elif A.ndim == 2:
        npidx = np.array(idx,copy=False)
        return npidx // A.shape[1], npidx % A.shape[1]
    else:
        raise ValueError("Dimension mismatch. Array should be either 1D or 2D")


@boundscheck(False)
@wraparound(False)
def findequal_approx(np.ndarray A, Real_t num, Real_t tol=1.0e-14):
    """An equivalent function to numpy.where(A==num). The difference here is that
        numpy.where results in a boolean array of size equal to the original
        array A, even if number of counts is none or a single element. This function
        allocates and does push_back on per count"""

    if A.dtype != Real:
        raise TypeError("Type mismatch. Array and value must have the same type")

    cdef Real_t[::1] arr = A.ravel()
    cdef vector[Integer_t] idx = FindEqualApprox(&arr[0], A.size, num, tol)

    if A.ndim == 1:
        return np.array(idx,copy=False)
    elif A.ndim == 2:
        npidx = np.array(idx,copy=False)
        return npidx // A.shape[1], npidx % A.shape[1]
    else:
        raise ValueError("Dimension mismatch. Array should be either 1D or 2D")


@boundscheck(False)
@wraparound(False)
def findless(np.ndarray A, Real_t num):
    """An equivalent function to numpy.where(A<num). The difference here is that
        numpy.where results in a boolean array of size equal to the original
        array A, even if number of counts is none or a single element. This function
        allocates and performs push_back per hit count"""

    cdef Real_t[::1] arr = A.ravel()
    cdef vector[Integer_t] idx = FindLessThan(&arr[0],A.size,num)

    if A.ndim == 1:
        return np.array(idx,copy=False)
    elif A.ndim == 2:
        npidx = np.array(idx,copy=False)
        return npidx // A.shape[1], npidx % A.shape[1]
    else:
        raise ValueError("Dimension mismatch. Array should be either 1D or 2D")


@boundscheck(False)
@wraparound(False)
def findgreater(np.ndarray A, Real_t num):
    """An equivalent function to numpy.where(A>num). The difference here is that
        numpy.where results in a boolean array of size equal to the original
        array A, even if number of counts is none or a single element. This function
        allocates and does push_back on per count"""

    cdef Real_t[::1] arr = A.ravel()
    cdef vector[Integer_t] idx = FindGreaterThan(&arr[0],A.size,num)

    if A.ndim == 1:
        return np.array(idx,copy=False)
    elif A.ndim == 2:
        npidx = np.array(idx,copy=False)
        return npidx // A.shape[1], npidx % A.shape[1]
    else:
        raise ValueError("Dimension mismatch. Array should be either 1D or 2D")





#cython: profile=False
#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False

import numpy as np
cimport numpy as np

ctypedef double Real


cdef extern from "_Piezoelectric_100_.h" nogil:
    cdef cppclass _Piezoelectric_100_[Real]:
        _Piezoelectric_100_() except +
        _Piezoelectric_100_(Real mu1, Real mu2, Real mu3, Real lamb, Real eps_1, Real eps_2, Real eps_3) except +
        void SetParameters(Real mu1, Real mu2, Real mu3, Real lamb, Real eps_1, Real eps_2, Real eps_3) except +
        void KineticMeasures(Real *Dnp, Real *Snp, Real* Hnp, int ndim, int ngauss, const Real *Fnp, const Real *Enp, const Real *Nnp) except +



def KineticMeasures(material, np.ndarray[Real, ndim=3, mode='c'] F, np.ndarray[Real, ndim=2] E,
    np.ndarray[Real, ndim=2] N):
    
    cdef int ndim = F.shape[2]
    cdef int ngauss = F.shape[0]
    cdef np.ndarray[Real, ndim=3, mode='c'] D, stress, hessian

    D = np.zeros((ngauss,ndim,1),dtype=np.float64)
    stress = np.zeros((ngauss,ndim,ndim),dtype=np.float64)
    if ndim==3:
        hessian = np.zeros((ngauss,9,9),dtype=np.float64)
    elif ndim==2:
        hessian = np.zeros((ngauss,5,5),dtype=np.float64)

    cdef _Piezoelectric_100_[Real] mat_obj = _Piezoelectric_100_()
    mat_obj.SetParameters(material.mu1,material.mu2,material.mu3,material.lamb,material.eps_1,material.eps_2,material.eps_3)
    mat_obj.KineticMeasures(&D[0,0,0], &stress[0,0,0], &hessian[0,0,0], ndim, ngauss, &F[0,0,0], &E[0,0], &N[0,0])

    return D, stress, hessian









# cdef class Piezoelectric_100:

#     # CREATE A POINTER TO CPP BASE CLASS
#     cdef _Piezoelectric_100_[Real] *baseptr
#     cdef public dict material

#     def __cinit__(self, material):

#         # CREATE A NEW CPP OBJECT BY CALLING ITS CONSTRUCTOR
#         self.baseptr = new _Piezoelectric_100_()
#         # CHECK IF THE OBJECT WAS CREATED
#         if self.baseptr is NULL:
#             raise MemoryError("Could not create an instance of {} material" % type(self).__name__)

#         self.baseptr.SetParameters(material.mu1,material.mu2,material.mu3,material.lamb,material.eps_1,material.eps_2,material.eps_3)
#         self.material = material.__dict__

#     def KineticMeasures(self,np.ndarray[Real, ndim=3, mode='c'] F, np.ndarray[Real, ndim=2] E,
#         np.ndarray[Real, ndim=2] N):
        
#         cdef int ndim = F.shape[2]
#         cdef int ngauss = F.shape[0]
#         cdef np.ndarray[Real, ndim=3, mode='c'] D, stress, hessian

#         if ndim==3:
#             D = np.zeros((ngauss,3,1),dtype=np.float64)
#             stress = np.zeros((ngauss,3,3),dtype=np.float64)
#             hessian = np.zeros((ngauss,9,9),dtype=np.float64)
#         elif ndim==2:
#             D = np.zeros((ngauss,2,1),dtype=np.float64)
#             stress = np.zeros((ngauss,2,2),dtype=np.float64)
#             hessian = np.zeros((ngauss,5,5),dtype=np.float64)

#         self.baseptr.KineticMeasures(&D[0,0,0], &stress[0,0,0], &hessian[0,0,0], ndim, ngauss, &F[0,0,0], &E[0,0], &N[0,0])

#         return D, stress, hessian


#     def __dealloc__(self):
#         if self.baseptr != NULL:
#             del self.baseptr



import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt

#####################################################################################################
                                # Anisotropic MooneyRivlin Model
#####################################################################################################


class AnisotropicMooneyRivlin_0(Material):
    """A compressible transervely isotropic model with the isotropic part being Mooney-Rivlin
        The energy is given by:

            W(C) =  gamma * ( alpha*(C:I) + beta*(G:I) ) +
                    eta*(1-alpha)*( (N C N)**2 + N G N) - ut*J + lambda/2*(J-1)**2

            ut = 2.*gamma*(alpha+2.0*beta) + 2.*(1. - gamma)*eta  # for the stress to be
                zero at the origin

        the parameter "gamma" controls the amount of anisotropy and the vector N(ndim,1) is
        the direction of anisotropy

    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(AnisotropicMooneyRivlin_0, self).__init__(mtype, ndim, **kwargs)
        self.nvar = self.ndim
        self.is_transversely_isotropic = True
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"

        if self.ndim==3:
            self.H_VoigtSize = 6
        else:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False


    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem][:,None]
        FN = np.dot(F,N)[:,0]
        HN = np.dot(H,N)[:,0]
        innerHN = einsum('i,i',HN,HN)
        outerHN = einsum('i,j',HN,HN)

        # FIX GAMMA
        gamma = 0.5
        # gamma = 1.0
        alpha = mu/2./gamma
        beta  = mu/2./gamma
        eta   = mu/3.
        ut    = 2.*gamma*(alpha+2.0*beta) + 2.*(1. - gamma)*eta
        lamb  = lamb + 2.*gamma*alpha - 2*(1.- gamma)*eta


        H_Voigt = 2.*gamma*beta/J* ( 2.0*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b) ) + \
                4.*(1-gamma)*eta/J * einsum('i,j,k,l',FN,FN,FN,FN) + \
                4.*(1-gamma)*eta/J * ( innerHN * einsum('ij,kl',I,I) - \
                0.5*innerHN * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) - \
                einsum('ij,k,l',I,HN,HN) - einsum('i,j,kl',HN,HN,I) ) + \
                2.*(1-gamma)*eta/J * ( einsum('il,j,k',I,HN,HN) + einsum('jl,i,k',I,HN,HN) + \
                einsum('ik,j,l',I,HN,HN) + einsum('jk,i,l',I,HN,HN) ) - \
                ut*einsum('ij,kl',I,I) + ut * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) + \
                lamb*(2.*J-1.)*einsum('ij,kl',I,I) - lamb*(J-1.) * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) )

        H_Voigt = Voigt(H_Voigt ,1)

        self.H_VoigtSize = H_Voigt.shape[0]

        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem][:,None]
        FN = np.dot(F,N)
        HN = np.dot(H,N)[:,0]
        innerHN = einsum('i,i',HN,HN)
        outerHN = einsum('i,j',HN,HN)

        mu = self.mu
        lamb = self.lamb

        # FIX GAMMA
        gamma = 0.5
        # gamma = 1.0
        alpha = mu/2./gamma
        beta  = mu/2./gamma
        eta   = mu/3.
        ut    = 2.*gamma*(alpha+2.0*beta) + 2.*(1. - gamma)*eta
        lamb  = lamb + 2.*gamma*alpha - 2*(1.- gamma)*eta


        if self.ndim == 3:
            trb = trace(b)
        elif self.ndim == 2:
            trb = trace(b) + 1


        stress = 2.*gamma*alpha/J*b + 2.*gamma*beta/J*(trb*b - np.dot(b,b)) + \
                 2.*(1.- gamma)*eta/J*np.dot(FN.T,FN)[0][0]*np.dot(FN,FN.T) + \
                 2.*(1.- gamma)*eta/J*(innerHN*I - outerHN) - \
                 ut*I + lamb*(J-1.)*I

        return stress


    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        ndim = StrainTensors['I'].shape[0]
        return np.zeros((ndim,1))





import numpy as np
from numpy import einsum
from Florence.Tensor import trace, Voigt
from .MaterialBase import Material
from Florence.LegendreTransform import LegendreTransform
from math import sqrt

class Multi_Piezoelectric_100(Material):
    """ 
                Piezoelectric model in terms of internal energy 
                W(C,D) = W_mn(C) + 1/2/eps_1 (D0*D0) + 1/2/eps_2/J (FD0*FD0) 
                    + u3*(FD0/sqrt(u3*eps_3)+FN)*(FD0/sqrt(u3*eps_3)+FN) + u3*HN*HN - 2*sqrt(u3/eps_3)*D0*N
                W_mn(C) = u1*C:I+u2*G:I - 2*(u1+2*u2+u3)*lnJ + lamb/2*(J-1)**2 

    """
    
    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(Multi_Piezoelectric_100, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.nvar = self.ndim+1
        self.energy_type = "internal_energy"
        self.legendre_transform = LegendreTransform()
        self.nature = "nonlinear"
        self.fields = "electro_mechanics"
        
        self.is_transversely_isotropic = True
        if self.ndim==3:
            self.H_VoigtSize = 9
        else:
            self.H_VoigtSize = 5

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx, elem=0):
        self.mu1 = self.mu1s[elem]
        self.mu2 = self.mu2s[elem]
        self.mu3 = self.mu3s[elem]
        self.lamb = self.lambs[elem]
        self.eps_1 = self.eps_1s[elem]
        self.eps_2 = self.eps_2s[elem]
        self.eps_3 = self.eps_3s[elem]

        from Florence.MaterialLibrary.LLDispatch._Piezoelectric_100_ import KineticMeasures
        return KineticMeasures(self,np.ascontiguousarray(F), ElectricFieldx, self.anisotropic_orientations[elem][:,None])

    def Hessian(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu1 = self.mu1s[elem]
        mu2 = self.mu2s[elem]
        mu3 = self.mu3s[elem]
        lamb = self.lambs[elem]
        eps_1 = self.eps_1s[elem]
        eps_2 = self.eps_2s[elem]
        eps_3 = self.eps_3s[elem]

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem][:,None]
        D  = ElectricDisplacementx.reshape(self.ndim,1)

        FN = np.dot(F,N)[:,0]
        HN = np.dot(H,N)[:,0]
        innerHN = einsum('i,i',HN,HN)
        outerHN = einsum('i,j',HN,HN)
        Dx = D.reshape(self.ndim)
        DD = np.dot(D.T,D)[0,0]

        # Iso + Aniso
        C_mech = 2.*mu2/J* ( 2.0*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b) ) + \
                2.*(mu1+2*mu2+mu3)/J * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) + \
                lamb*(2.*J-1.)*einsum('ij,kl',I,I) - lamb*(J-1.) * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) - \
                4.*mu3/J * ( einsum('ij,kl',I,outerHN) + einsum('ij,kl',outerHN,I) ) + \
                2.*mu3/J*innerHN*(2.0*einsum('ij,kl',I,I) - einsum('ik,jl',I,I) - einsum('il,jk',I,I) ) + \
                2.*mu3/J * ( einsum('il,j,k',I,HN,HN) + einsum('jl,i,k',I,HN,HN) + \
                einsum('ik,j,l',I,HN,HN) + einsum('jk,i,l',I,HN,HN) )

        C_elect = 1./eps_2*(0.5*DD*(einsum('ik,jl',I,I) + einsum('il,jk',I,I) + einsum('ij,kl',I,I) ) - \
                einsum('ij,k,l',I,Dx,Dx) - einsum('i,j,kl',Dx,Dx,I)) 

        self.elasticity_tensor = C_mech + C_elect

        
        self.coupling_tensor = 1./eps_2*(einsum('ik,j',I,Dx) + einsum('i,jk',Dx,I) - einsum('ij,k',I,Dx)) + \
                2.*J*sqrt(mu3/eps_3)*(einsum('ik,j',I,Dx) + einsum('i,jk',Dx,I)) + \
                2.*sqrt(mu3/eps_3)*(einsum('ik,j',I,FN) + einsum('i,jk',FN,I))


        self.dielectric_tensor = J/eps_1*np.linalg.inv(b)  + 1./eps_2*I + 2.*J*sqrt(mu3/eps_3)*I

        # TRANSFORM TENSORS TO THEIR ENTHALPY COUNTERPART
        E_Voigt, P_Voigt, C_Voigt = self.legendre_transform.InternalEnergyToEnthalpy(self.dielectric_tensor, 
            self.coupling_tensor, self.elasticity_tensor)

        # BUILD HESSIAN
        factor = -1.
        H1 = np.concatenate((C_Voigt,factor*P_Voigt),axis=1)
        H2 = np.concatenate((factor*P_Voigt.T,E_Voigt),axis=1)
        H_Voigt = np.concatenate((H1,H2),axis=0)



        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu1 = self.mu1s[elem]
        mu2 = self.mu2s[elem]
        mu3 = self.mu3s[elem]
        lamb = self.lambs[elem]
        eps_1 = self.eps_1s[elem]
        eps_2 = self.eps_2s[elem]
        eps_3 = self.eps_3s[elem]

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem][:,None]
        FN = np.dot(F,N)[:,0]
        HN = np.dot(H,N)[:,0]
        outerFN = einsum('i,j',FN,FN)
        innerHN = einsum('i,i',HN,HN)
        outerHN = einsum('i,j',HN,HN)

        D  = ElectricDisplacementx.reshape(self.ndim,1)
        Dx = D.reshape(self.ndim)
        DD = np.dot(D.T,D)[0,0]
        D0D = np.dot(D,D.T)

        if self.ndim == 3:
            trb = trace(b)
        elif self.ndim == 2:
            trb = trace(b) + 1.

        sigma_mech = 2.*mu1/J*b + \
            2.*mu2/J*(trb*b - np.dot(b,b)) - \
            2.*(mu1+2*mu2+mu3)/J*I + \
            lamb*(J-1)*I +\
            2*mu3/J*outerFN +\
            2*mu3/J*innerHN*I - 2*mu3/J*outerHN

        sigma_electric = 1./eps_2*(D0D - 0.5*DD*I) +\
            2.*J*sqrt(mu3/eps_3)*D0D + 2*sqrt(mu3/eps_3)*(einsum('i,j',Dx,FN) + einsum('i,j',FN,Dx))

        sigma = sigma_mech + sigma_electric

        return sigma


    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        # THE ELECTRIC FIELD NEEDS TO BE MODFIED TO TAKE CARE OF CONSTANT TERMS
        mu1 = self.mu1s[elem]
        mu2 = self.mu2s[elem]
        mu3 = self.mu3s[elem]
        lamb = self.lambs[elem]
        eps_1 = self.eps_1s[elem]
        eps_2 = self.eps_2s[elem]
        eps_3 = self.eps_3s[elem]

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem][:,None]
        FN = np.dot(F,N)
        HN = np.dot(H,N)
        E = ElectricFieldx.reshape(self.ndim,1)
        modElectricFieldx = (E - 2.*sqrt(mu3/eps_3)*FN + 2./J*sqrt(mu3/eps_3)*HN) 

        # D = self.legendre_transform.GetElectricDisplacement(self, StrainTensors, modElectricFieldx, elem, gcounter)
        
        # SANITY CHECK FOR IMPLICIT COMPUTATUTAION OF D
        inverse = np.linalg.inv(J/eps_1*np.linalg.inv(b) + 1./eps_2*I + 2.*J*sqrt(mu3/eps_3)*I)
        D_exact = np.dot(inverse, (E - 2.*sqrt(mu3/eps_3)*FN + 2./J*sqrt(mu3/eps_3)*HN) ) 
        # print np.linalg.norm(D - D_exact)
        return D_exact

        return D



import numpy as np
from .MaterialBase import Material
from Florence.Tensor import trace


class NeoHookean_1(Material):
    """NeoHookean model with the following energy

        W(C) = u/2*C:I -u*J + lambda *(J-1)**2

        """
    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(NeoHookean_1, self).__init__(mtype, ndim, **kwargs)

        self.is_transversely_isotropic = False
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"  

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False


    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):
        mu = self.mu
        lamb = self.lamb
        I = StrainTensors['I']
        detF = StrainTensors['J'][gcounter]

        mu2 = mu - lamb*(detF-1.0)
        lamb2 = lamb*(2*detF-1.0) - mu

        C_Voigt = lamb2*self.vIijIkl+mu2*self.vIikIjl

        self.H_VoigtSize = C_Voigt.shape[0]

        return C_Voigt


    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        mu = self.mu
        lamb = self.lamb

        return (lamb*(J-1.0)-mu)*I+1.0*mu/J*b





#cython: profile=False
#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False

import numpy as np
cimport numpy as np

ctypedef double Real

cdef extern from "_ConstitutiveStiffnessDF_.h":
    
    inline void _ConstitutiveStiffnessIntegrandDF_Filler_(Real *stiffness, Real *traction,
        const Real* SpatialGradient,
        const Real* CauchyStressTensor,
        const Real* H_Voigt,
        const Real* detJ,
        int ngauss, 
        int noderpelem, 
        int ndim, 
        int nvar, 
        int H_VoigtSize,
        int requires_geometry_update) nogil


def __ConstitutiveStiffnessIntegrandDF__(np.ndarray[Real, ndim=3, mode='c'] SpatialGradient,
    np.ndarray[Real, ndim=3, mode='c'] CauchyStressTensor,
    np.ndarray[Real, ndim=3, mode='c'] H_Voigt,
    np.ndarray[Real, ndim=1] detJ, 
    int nvar,
    int requires_geometry_update):

    cdef int ngauss = SpatialGradient.shape[0]
    cdef int nodeperelem = SpatialGradient.shape[1]
    cdef int ndim = SpatialGradient.shape[2]
    cdef int local_size = nvar*nodeperelem
    cdef int H_VoigtSize = H_Voigt.shape[2]

    cdef np.ndarray[Real, ndim=2, mode='c'] traction = np.zeros((local_size,1),dtype=np.float64)
    cdef np.ndarray[Real, ndim=2, mode='c'] stiffness = np.zeros((local_size,
        local_size),dtype=np.float64)

    _ConstitutiveStiffnessIntegrandDF_Filler_(&stiffness[0,0], &traction[0,0],
        &SpatialGradient[0,0,0], 
        &CauchyStressTensor[0,0,0], 
        &H_Voigt[0,0,0], &detJ[0],
        ngauss,
        nodeperelem,
        ndim,
        nvar,
        H_VoigtSize,
        requires_geometry_update)

    return stiffness, traction


import numpy as np
from numpy import einsum
from Florence.Tensor import trace, Voigt
from .MaterialBase import Material
from Florence.LegendreTransform import LegendreTransform


class AnisotropicIdealDielectric(Material):
    """Ideal dielectric material model for Laplace problems
    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(AnisotropicIdealDielectric, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.nvar = 1
        self.energy_type = "internal_energy"
        self.legendre_transform = LegendreTransform()
        self.nature = "linear"
        self.fields = "electrostatics"

        self.H_VoigtSize = self.ndim

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self, F, ElectricFieldx, elem=0):
        # from Florence.MaterialLibrary.LLDispatch._AnisotropicIdealDielectric_ import KineticMeasures
        # D, H_Voigt = KineticMeasures(self,np.ascontiguousarray(F), ElectricFieldx)
        # return D, None, H_Voigt
        ndim = self.ndim
        e = self.e # dielectric tensor
        J = np.linalg.det(F)
        I = np.eye(ndim,ndim)
        et = np.tile(e,(ElectricFieldx.shape[0],1)).reshape(ElectricFieldx.shape[0],e.shape[0],e.shape[0])
        eJ = np.einsum("ijk,i->ijk",et,1./J)
        D = np.einsum("ijk,ik->ij",eJ,ElectricFieldx)[...,None]
        # Negative definite inverse is needed due to Legendre transform
        H_Voigt = -eJ

        return np.ascontiguousarray(D), None, np.ascontiguousarray(H_Voigt)

    def Hessian(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        e = self.e
        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        # self.dielectric_tensor = 1./eps_1*I
        # Negative definite inverse is needed due to Legendre transform
        self.dielectric_tensor = -e/J
        return self.dielectric_tensor

    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        pass

    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        # D = self.legendre_transform.GetElectricDisplacement(self, StrainTensors, ElectricFieldx, elem, gcounter)
        e = self.e
        J = StrainTensors['J'][gcounter]
        E = ElectricFieldx.reshape(self.ndim,1)
        D_exact = np.dot(e/J,E)
        return D_exact

    def Permittivity(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        e = self.e
        J = StrainTensors['J'][gcounter]
        # Negative definite inverse is needed due to Legendre transform
        self.dielectric_tensor = -e/J
        return self.dielectric_tensor

    def ElectrostaticMeasures(self,F,ElectricFieldx, elem=0):
        return self.KineticMeasures(F,ElectricFieldx, elem)


import numpy as np
from .VariationalPrinciple import VariationalPrinciple
from Florence import QuadratureRule, FunctionSpace

from Florence.FiniteElements.LocalAssembly.KinematicMeasures import *
from Florence.FiniteElements.LocalAssembly._KinematicMeasures_ import _KinematicMeasures_
from ._ConstitutiveStiffnessLaplacian_ import __ConstitutiveStiffnessIntegrandLaplacian__
from Florence.Tensor import issymetric
from Florence.LegendreTransform import LegendreTransform

__all__ = ["LaplacianFormulation"]

class LaplacianFormulation(VariationalPrinciple):

    def __init__(self, mesh, variables_order=(1,),
        quadrature_rules=None, quadrature_type=None, function_spaces=None, compute_post_quadrature=True,
        equally_spaced_bases=False):

        if mesh.element_type != "tet" and mesh.element_type != "tri" and \
            mesh.element_type != "quad" and mesh.element_type != "hex":
            raise NotImplementedError( type(self).__name__, "has not been implemented for", mesh.element_type, "elements")

        if isinstance(variables_order,int):
            self.variables_order = (self.variables_order,)
        self.variables_order = variables_order

        super(LaplacianFormulation, self).__init__(mesh,variables_order=self.variables_order,
            quadrature_type=quadrature_type,quadrature_rules=quadrature_rules,function_spaces=function_spaces,
            compute_post_quadrature=compute_post_quadrature)

        self.fields = "electrostatics"
        self.nvar = 1

        C = mesh.InferPolynomialDegree() - 1
        mesh.InferBoundaryElementType()

        if quadrature_rules == None and self.quadrature_rules == None:

            # OPTION FOR QUADRATURE TECHNIQUE FOR TRIS AND TETS
            optimal_quadrature = 3
            # is_flattened = True
            is_flattened = False

            if mesh.element_type == "tri" or mesh.element_type == "tet":
                norder = 2*C
                # TAKE CARE OF C=0 CASE
                if norder == 0:
                    norder = 1

                norder_post = 2*(C+1)
            else:
                norder = C+2
                norder_post = 2*(C+2)

            # GET QUADRATURE
            quadrature = QuadratureRule(optimal=optimal_quadrature, norder=norder, mesh_type=mesh.element_type, is_flattened=is_flattened)
            if self.compute_post_quadrature:
                # COMPUTE INTERPOLATION FUNCTIONS AT ALL INTEGRATION POINTS FOR POST-PROCESSING
                post_quadrature = QuadratureRule(optimal=optimal_quadrature, norder=norder_post, mesh_type=mesh.element_type)
            else:
                post_quadrature = None

            # BOUNDARY QUADRATURE
            bquadrature = QuadratureRule(optimal=optimal_quadrature, norder=C+2, mesh_type=mesh.boundary_element_type, is_flattened=is_flattened)

            self.quadrature_rules = (quadrature,post_quadrature,bquadrature)
        else:
            self.quadrature_rules = quadrature_rules

        if function_spaces == None and self.function_spaces == None:

            # CREATE FUNCTIONAL SPACES
            function_space = FunctionSpace(mesh, quadrature, p=C+1, equally_spaced=equally_spaced_bases, use_optimal_quadrature=is_flattened)
            if compute_post_quadrature:
                post_function_space = FunctionSpace(mesh, post_quadrature, p=C+1, equally_spaced=equally_spaced_bases)
            else:
                post_function_space = None

            # CREATE BOUNDARY FUNCTIONAL SPACES
            bfunction_space = FunctionSpace(mesh.CreateDummyLowerDimensionalMesh(),
                bquadrature, p=C+1, equally_spaced=equally_spaced_bases, use_optimal_quadrature=is_flattened)

            self.function_spaces = (function_space,post_function_space,bfunction_space)
        else:
            self.function_spaces = function_spaces


        # local_size = function_space.Bases.shape[0]*self.nvar
        local_size = self.function_spaces[0].Bases.shape[0]*self.nvar
        self.local_rows = np.repeat(np.arange(0,local_size),local_size,axis=0)
        self.local_columns = np.tile(np.arange(0,local_size),local_size)
        self.local_size = local_size

        # FOR MASS
        local_size_m = self.function_spaces[0].Bases.shape[0]*self.nvar
        self.local_rows_mass = np.repeat(np.arange(0,local_size_m),local_size_m,axis=0)
        self.local_columns_mass = np.tile(np.arange(0,local_size_m),local_size_m)
        self.local_size_m = local_size_m


    def GetElementalMatrices(self, elem, function_space, mesh, material, fem_solver, Eulerx, Eulerp):

        massel=[]; f = []
        # GET THE FIELDS AT THE ELEMENT LEVEL
        LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
        EulerElemCoords = Eulerx[mesh.elements[elem,:],:]
        ElectricPotentialElem = Eulerp[mesh.elements[elem,:]]

        # COMPUTE THE STIFFNESS MATRIX
        if material.has_low_level_dispatcher:
            stiffnessel, t = self.__GetLocalStiffness__(function_space, material, LagrangeElemCoords,
                EulerElemCoords, ElectricPotentialElem, fem_solver, elem)
        else:
            stiffnessel, t = self.GetLocalStiffness(function_space, material, LagrangeElemCoords,
                EulerElemCoords, ElectricPotentialElem, fem_solver, elem)


        if fem_solver.has_moving_boundary:
            # COMPUTE FORCE VECTOR
            f = ApplyNeumannBoundaryConditions3D(MainData, mesh, elem, LagrangeElemCoords)

        I_stiff_elem, J_stiff_elem, V_stiff_elem = self.FindIndices(stiffnessel)

        I_mass_elem = []; J_mass_elem = []; V_mass_elem = []
        return I_stiff_elem, J_stiff_elem, V_stiff_elem, t, f, I_mass_elem, J_mass_elem, V_mass_elem


    def GetElementalMatricesInVectorForm(self, elem, function_space, mesh, material, fem_solver, Eulerx, Eulerp):

        massel=[]; f = []
        # GET THE FIELDS AT THE ELEMENT LEVEL
        LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
        EulerElemCoords = Eulerx[mesh.elements[elem,:],:]
        ElectricPotentialElem = Eulerp[mesh.elements[elem,:]]

        # COMPUTE THE TRACTION VECTOR
        if material.has_low_level_dispatcher:
            t = self.__GetLocalTraction__(function_space, material, LagrangeElemCoords,
                EulerElemCoords, ElectricPotentialElem, fem_solver, elem)
        else:
            t = self.GetLocalTraction(function_space, material, LagrangeElemCoords,
                EulerElemCoords, ElectricPotentialElem, fem_solver, elem)

        if fem_solver.has_moving_boundary:
            # COMPUTE FORCE VECTOR
            f = self.ApplyNeumannBoundaryConditions3D(fem_solver, mesh, elem, LagrangeElemCoords, ElectricPotentialElem)

        return t, f, massel



    def GetLocalStiffness(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        """Get stiffness matrix of the system"""

        nvar = self.nvar
        ndim = self.ndim
        nodeperelem = function_space.Bases.shape[0]

        det = np.linalg.det
        inv = np.linalg.inv
        Jm = function_space.Jm
        AllGauss = function_space.AllGauss

        # ALLOCATE
        stiffness = np.zeros((nodeperelem*nvar,nodeperelem*nvar),dtype=np.float64)
        tractionforce = np.zeros((nodeperelem*nvar,1),dtype=np.float64)
        B = np.zeros((nodeperelem*nvar,material.H_VoigtSize),dtype=np.float64)

        # COMPUTE KINEMATIC MEASURES AT ALL INTEGRATION POINTS USING EINSUM (AVOIDING THE FOR LOOP)
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientX = np.einsum('ijk,jl->kil', Jm, LagrangeElemCoords)
        # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        MaterialGradient = np.einsum('ijk,kli->ijl', inv(ParentGradientX), Jm)
        # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
        F = np.einsum('ij,kli->kjl', EulerELemCoords, MaterialGradient)

        # COMPUTE REMAINING KINEMATIC MEASURES
        StrainTensors = KinematicMeasures(F, fem_solver.analysis_nature)

        # UPDATE/NO-UPDATE GEOMETRY
        if fem_solver.requires_geometry_update:
            # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
            ParentGradientx = np.einsum('ijk,jl->kil',Jm,EulerELemCoords)
            # SPATIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla (N)]
            SpatialGradient = np.einsum('ijk,kli->ilj',inv(ParentGradientx),Jm)
            # COMPUTE ONCE detJ (GOOD SPEEDUP COMPARED TO COMPUTING TWICE)
            detJ = np.einsum('i,i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)),np.abs(StrainTensors['J']))
        else:
            # SPATIAL GRADIENT AND MATERIAL GRADIENT TENSORS ARE EQUAL
            SpatialGradient = np.einsum('ikj',MaterialGradient)
            # COMPUTE ONCE detJ
            detJ = np.einsum('i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)))

        # GET ELECTRIC FIELD
        ElectricFieldx = - np.einsum('ijk,j',SpatialGradient,ElectricPotentialElem)

        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            if material.energy_type == "enthalpy":
                # COMPUTE THE HESSIAN AT THIS GAUSS POINT
                H_Voigt = material.Hessian(StrainTensors,ElectricFieldx[counter,:], elem, counter)

                # COMPUTE ELECTRIC DISPLACEMENT
                ElectricDisplacementx = material.ElectricDisplacementx(StrainTensors, ElectricFieldx[counter,:], elem, counter)

            elif material.energy_type == "internal_energy":
                # THIS REQUIRES LEGENDRE TRANSFORM

                # COMPUTE ELECTRIC DISPLACEMENT IMPLICITLY
                ElectricDisplacementx = material.ElectricDisplacementx(StrainTensors, ElectricFieldx[counter,:], elem, counter)

                # COMPUTE THE HESSIAN AT THIS GAUSS POINT
                H_Voigt = material.Hessian(StrainTensors,ElectricDisplacementx, elem, counter)

            # COMPUTE THE TANGENT STIFFNESS MATRIX
            BDB_1, t = self.ConstitutiveStiffnessIntegrand(B, SpatialGradient[counter,:,:],
                ElectricDisplacementx, H_Voigt, analysis_nature=fem_solver.analysis_nature,
                has_prestress=fem_solver.has_prestress)

            # INTEGRATE TRACTION FORCE
            tractionforce += t*detJ[counter]
            # INTEGRATE STIFFNESS
            stiffness += BDB_1*detJ[counter]

        return stiffness, tractionforce


    def __GetLocalStiffness__(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        """Get stiffness matrix of the system"""

        # GET LOCAL KINEMATICS
        SpatialGradient, F, detJ = _KinematicMeasures_(function_space.Jm, function_space.AllGauss[:,0], LagrangeElemCoords,
            EulerELemCoords, fem_solver.requires_geometry_update)
        # GET ELECTRIC FIELD
        ElectricFieldx = - np.einsum('ijk,j',SpatialGradient,ElectricPotentialElem)
        # COMPUTE WORK-CONJUGATES AND HESSIAN AT THIS GAUSS POINT
        ElectricDisplacementx, _, H_Voigt = material.KineticMeasures(F, ElectricFieldx, elem=elem)
        # COMPUTE LOCAL CONSTITUTIVE STIFFNESS AND TRACTION
        stiffness, tractionforce = __ConstitutiveStiffnessIntegrandLaplacian__(SpatialGradient,ElectricDisplacementx,
            H_Voigt,detJ,self.nvar,fem_solver.requires_geometry_update)

        return stiffness, tractionforce


    def ConstitutiveStiffnessIntegrand(self, B, SpatialGradient, ElectricDisplacementx,
            H_Voigt, analysis_nature="nonlinear", has_prestress=True):

            ndim = self.ndim
            nvar = self.nvar

            # MATRIX FORM
            SpatialGradient = SpatialGradient.T
            TotalTraction = ElectricDisplacementx.reshape(ElectricDisplacementx.shape[0],1)

            # THREE DIMENSIONS
            if SpatialGradient.shape[0]==3:
                # Electrostatic
                B[::nvar,0] = SpatialGradient[0,:]
                B[::nvar,1] = SpatialGradient[1,:]
                B[::nvar,2] = SpatialGradient[2,:]

            elif SpatialGradient.shape[0]==2:
                # Electrostatic
                B[::nvar,0] = SpatialGradient[0,:]
                B[::nvar,1] = SpatialGradient[1,:]

            BDB = np.dot(np.dot(B,H_Voigt),B.T)
            t = np.dot(B,TotalTraction)

            return BDB, t



    def GetLocalTraction(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        """Get traction vector of the system"""

        nvar = self.nvar
        ndim = self.ndim
        nodeperelem = function_space.Bases.shape[0]

        det = np.linalg.det
        inv = np.linalg.inv
        Jm = function_space.Jm
        AllGauss = function_space.AllGauss

        # ALLOCATE
        tractionforce = np.zeros((nodeperelem*nvar,1),dtype=np.float64)
        B = np.zeros((nodeperelem*nvar,material.H_VoigtSize),dtype=np.float64)

        # COMPUTE KINEMATIC MEASURES AT ALL INTEGRATION POINTS USING EINSUM (AVOIDING THE FOR LOOP)
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientX = np.einsum('ijk,jl->kil', Jm, LagrangeElemCoords)
        # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        MaterialGradient = np.einsum('ijk,kli->ijl', inv(ParentGradientX), Jm)
        # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
        F = np.einsum('ij,kli->kjl', EulerELemCoords, MaterialGradient)

        # COMPUTE REMAINING KINEMATIC MEASURES
        StrainTensors = KinematicMeasures(F, fem_solver.analysis_nature)

        # UPDATE/NO-UPDATE GEOMETRY
        if fem_solver.requires_geometry_update:
            # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
            ParentGradientx = np.einsum('ijk,jl->kil',Jm,EulerELemCoords)
            # SPATIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla (N)]
            SpatialGradient = np.einsum('ijk,kli->ilj',inv(ParentGradientx),Jm)
            # COMPUTE ONCE detJ (GOOD SPEEDUP COMPARED TO COMPUTING TWICE)
            detJ = np.einsum('i,i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)),np.abs(StrainTensors['J']))
        else:
            # SPATIAL GRADIENT AND MATERIAL GRADIENT TENSORS ARE EQUAL
            SpatialGradient = np.einsum('ikj',MaterialGradient)
            # COMPUTE ONCE detJ
            detJ = np.einsum('i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)))

        # GET ELECTRIC FIELD
        ElectricFieldx = - np.einsum('ijk,j',SpatialGradient,ElectricPotentialElem)

        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            if material.energy_type == "enthalpy":

                # COMPUTE ELECTRIC DISPLACEMENT
                ElectricDisplacementx = material.ElectricDisplacementx(StrainTensors, ElectricFieldx[counter,:], elem, counter)

            elif material.energy_type == "internal_energy":
                # THIS REQUIRES LEGENDRE TRANSFORM

                # COMPUTE ELECTRIC DISPLACEMENT IMPLICITLY
                ElectricDisplacementx = material.ElectricDisplacementx(StrainTensors, ElectricFieldx[counter,:], elem, counter)

            # COMPUTE THE TANGENT STIFFNESS MATRIX
            t = self.TractionIntegrand(B, SpatialGradient[counter,:,:],
                ElectricDisplacementx, analysis_nature=fem_solver.analysis_nature,
                has_prestress=fem_solver.has_prestress)

            # INTEGRATE TRACTION FORCE
            tractionforce += t*detJ[counter]

        return tractionforce

    def __GetLocalTraction__(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        """Get traction vector of the system"""
        pass

    def TractionIntegrand(self, B, SpatialGradient, ElectricDisplacementx,
        analysis_nature="nonlinear", has_prestress=True):
        """Applies to displacement potential based formulation"""

        ndim = self.ndim
        nvar = self.nvar

        # MATRIX FORM
        SpatialGradient = SpatialGradient.T
        TotalTraction = ElectricDisplacementx.reshape(ElectricDisplacementx.shape[0],1)

        # THREE DIMENSIONS
        if SpatialGradient.shape[0]==3:
            # Electrostatic
            B[::nvar,0] = SpatialGradient[0,:]
            B[::nvar,1] = SpatialGradient[1,:]
            B[::nvar,2] = SpatialGradient[2,:]

        elif SpatialGradient.shape[0]==2:
            # Electrostatic
            B[::nvar,0] = SpatialGradient[0,:]
            B[::nvar,1] = SpatialGradient[1,:]

        t = np.dot(B,TotalTraction)

        return t


    def GetLocalResidual(self):
        pass


from __future__ import print_function
import os, sys


def PWD(filename=None):
    """Returns directory path of the current file"""
    if filename==None:
        filename = __file__
    return os.path.dirname(os.path.realpath(filename))

def RSWD():
    """Returns directory path of the file 
        from which the python script was run"""
    return os.path.dirname(sys.argv[0])


# FOR RunSession SCRIPTS - OLD
def SetPath(Pr,pwd,C,nelem,Analysis,AnalysisType, MaterialArgsType):
    
    # STORE PATHS FOR MAIN, CORE & PROBLEM DIRECTORIES
    ############################################################################
    class Path(object):
        """Getting directory paths"""

        TopLevel = pwd
        Main = pwd+'/Main/FiniteElements'
        Problem = os.path.dirname(Pr.__file__)
        Core = pwd+'/Core/FiniteElements'

    Path = Path


    if os.path.isdir(Path.Problem+'/Results'):
        print('Writing results in the problem directory:', Path.Problem)
    else:
        print('Writing the results in problem directory:', Path.Problem)
        os.mkdir(Path.Problem+'/Results')


    Path.ProblemResults = Path.Problem+'/Results/'
    Path.ProblemResultsFileNameMATLAB = 'Results_h'+str(nelem)+'_C'+str(C)+'.mat'
    # FOR NON-LINEAR ANALYSIS - DO NOT ADD THE EXTENSION
    Path.ProblemResultsFileNameVTK = 'Results_h'+str(nelem)+'_C'+str(C)
    # FOR LINEAR ANALYSIS
    # MainData.Path.ProblemResultsFileNameVTK = 'Results_h'+str(nelem)+'_C'+str(C)+'.vtu'

    # CONSIDERATION OF MATERAIL MODEL
    Path.MaterialModel = MaterialArgsType + '_Model/'

    # ANALYSIS SPECIFIC DIRECTORIES
    if Analysis == 'Static':
        if AnalysisType == 'Linear':
            Path.Analysis = 'LinearStatic/'     # ONE STEP/INCREMENT
        # MainData.Path.LinearDynamic = 'LinearDynamic'
        elif AnalysisType == 'Nonlinear':
            Path.Analysis = 'NonlinearStatic/'  # MANY INCREMENTS
        # Subdirectories
        if os.path.isdir(Path.ProblemResults+Path.Analysis):
            if not os.path.isdir(Path.ProblemResults+Path.Analysis+Path.MaterialModel):
                os.mkdir(Path.ProblemResults+Path.Analysis+Path.MaterialModel)
        else:
            os.mkdir(Path.ProblemResults+Path.Analysis)
            if not os.path.isdir(Path.ProblemResults+Path.Analysis+Path.MaterialModel):
                os.mkdir(Path.ProblemResults+Path.Analysis+Path.MaterialModel)

    elif Analysis == 'Dynamic':
        Path.Analysis = 'NonlinearDynamic/'
        # SUBDIRECTORIES
        if os.path.isdir(MainData.Path.ProblemResults+Path.Analysis):
            if not os.path.isdir(Path.ProblemResults+Path.Analysis+Path.MaterialModel):
                os.mkdir(Path.ProblemResults+Path.Analysis+Path.MaterialModel)
        else:
            os.mkdir(MainData.Path.ProblemResults+MainData.Path.Analysis)
            if not os.path.isdir(Path.ProblemResults+Path.Analysis+Path.MaterialModel):
                os.mkdir(Path.ProblemResults+Path.Analysis+Path.MaterialModel)


    return Path